from copy import deepcopy
from enum import Enum
from typing import List

from pycivil.EXAUtils.EXAExceptions import EXAExceptions as Ex
from pycivil.EXAStructural.codes import Code
from pycivil.EXAStructural.loads import ForcesOnSection


class Checkable:
    def __init__(self):
        self.__results = {}
        self.__ck = None
        self.__criteria = None
        self.__criteriaForCheck = []
        self.__code = None
        self.__loads = None
        self.__loadsSelector = None

    def _getResults(self):
        return self.__results

    def _setCode(self, code: Code):
        if not isinstance(code, Code):
            raise Ex("002", "code must be a Code type", type(code))
        self.__code = code

    def _setLoads(self, loads: List[ForcesOnSection]):
        if not isinstance(loads, list):
            raise Ex("001", "loads must be a list type", type(loads))
        self.__loads = loads

    def _setCriteria(self, criteria: List[str]):
        self.__criteria = criteria

    def _setCheckableObj(self, obj):
        self.__ck = obj

    def _setLoadsSelector(self, s):
        if not isinstance(s, list):
            raise Ex("001", "s must be a list type", type(s))
        self.__loadsSelector = s

    def _validateCriteria(self, criteria: List[str]):
        if not isinstance(criteria, list):
            raise Ex("0001", "criteria must be a List[str]", type(criteria))

        for v in criteria:
            if not isinstance(v, str):
                raise Ex("0001", "criteria must be a List[str]", type(v))
            assert self.__criteria is not None
            if v not in self.__criteria:
                raise Ex("0001", "value is not in criteria", v)

        self.__criteriaForCheck = criteria

    def _ck(self):
        return self.__ck

    def _criteriaForCheck(self):
        return self.__criteriaForCheck

    def check(
        self,
        criteria: List[str],
        loads: List[ForcesOnSection],
        law: Code,
        loadSelector: List[List[int]],
    ) -> dict:
        raise NotImplementedError("Need to be implemented")

    def getResults(self):
        return deepcopy(self.__results)


class Checker:
    def __init__(self):
        self.__checkable = None
        self.__results = None

    def _setResults(self, r):
        if isinstance(r, dict):
            self.__results = r
        else:
            raise TypeError("arguments must be a dict type")

    def check(
        self,
        criteria: List[str],
        loads: List[ForcesOnSection],
        law: Code,
        loadsSelector: List[List[int]] = (),
    ) -> None:
        if self.__checkable is None:
            raise ValueError("Before launch check need to use setCheckable() method")

        self.__results = self.__checkable.check(criteria, loads, law, loadsSelector)

    def setCheckable(self, c: Checkable):
        if not isinstance(c, Checkable):
            raise TypeError("arguments must be a checkable")

        self.__checkable = c

    def getResults(self) -> dict:
        return self.__results


class CheckableCriteriaEnum(str, Enum):
    SLE_NM = "SLE-NM"
    SLE_F = "SLE-F"
    SLU_NM = "SLU-NM"
    SLU_T = "SLU-T"
    SLU_NM_FIRE = "SLU-NM-FIRE"



import pytest
from pycivil.EXAUtils.solver import *


class FuckIn(BaseModel):
    pass


class FuckOut(BaseModel):
    pass


class FuckLogs(BaseModel):
    pass


class FuckRes(BaseModel):
    pass


def test_solver_init() -> None:
    solver = Solver(FuckIn())

    assert isinstance(solver.getModelInput(), FuckIn)
    assert solver.getModelOutput() is None
    assert solver.getModelLogs() is None

    with pytest.raises(NotImplementedError):
        solver.setModelInput(FuckIn())
    assert solver._setModelOutput(FuckOut())
    assert solver._setModelLogs(FuckLogs())
    assert solver._setModelResources(FuckRes())

    assert solver.setJobPath("fuck")
    assert solver.getJobPath() == "fuck"

    assert solver.getModelResources() == FuckRes()

    with pytest.raises(NotImplementedError):
        solver.run()

    with pytest.raises(NotImplementedError):
        solver.buildReport()

    with pytest.raises(NotImplementedError):
        solver.buildResources()

    with pytest.raises(NotImplementedError):
        solver.setModelInput(FuckIn())

    with pytest.raises(NotImplementedError):
        solver.exportModelInput()



import math
from enum import Enum
from typing import Literal, Tuple, Union

import numpy as np

from pycivil.EXAUtils import EXAExceptions as ex
from pycivil.EXAUtils.logging import log
from pycivil.EXAUtils.tables import Table


def fcm(fck: float) -> float:
    """Resistenza media a compressione nel calcestruzzo secondo NTC2018 [11.2.10.1]

    Args:
        fck (float): Resistenza cilindrica caratteristica del calcestruzzo

    Raises:
        ex.EXAExceptions: args fck must be float
        ex.EXAExceptions: args fck must be > 0

    Returns:
        float: resistenza media a compressione
    """
    if type(fck) != float and type(fck) != int:
        raise ex.EXAExceptions("(fcm)-0001", "args fck must be float or int", type(fck))

    if fck < 0.0:
        raise ex.EXAExceptions("(fcm)-0002", "args fck must be > 0", fck)

    _fcm = fck + 8

    return _fcm


def fctm(fck: Union[float, int]):
    """Resistenza media a trazione nel calcestruzzo secondo NTC2018 [11.2.10.2]

    Restituisce il valore della resistenza media a trazione nel calcestruzzo
    secondo NTC2018.

    Unità in MPa

    Args:
        fck (float): Resistenza cilindrica caratteristica del calcestruzzo

    Returns:
        fctm (float): resistenza media a trazione
    """
    if type(fck) not in (float, int):
        raise ex.EXAExceptions(
            "(fctm)-0001", "args fck must be float or int", type(fck)
        )

    if fck < 0.0:
        raise ex.EXAExceptions("(fctm)-0002", "args fck must be > 0", fck)

    if fck <= 50:
        _fctm = 0.30 * math.pow(fck, 0.666)
    else:
        _fctm = 2.12 * math.log(1 + fcm(fck) / 10)

    return _fctm


def fcfk(fck: Union[float, int]) -> float:
    """Resistenza caratteristica a trazione per flessione NTC2018 [11.2.10.2]

    Args:
        fck (Union(float,int)): Resistenza cilindrica caratteristica del calcestruzzo

    Raises:
        ex.EXAExceptions: args fck must be float or int
        ex.EXAExceptions: args fck must be > 0

    Returns:
        float: fcfk
    """
    if type(fck) not in (float, int):
        raise ex.EXAExceptions("0001", "args fck must be float or int", type(fck))

    if fck < 0.0:
        raise ex.EXAExceptions("0002", "args fck must be > 0", fck)

    return 0.7 * 1.2 * fctm(fck)


def Ecm(fck):
    """Modulo elastico istantaneo del calcestruzzo secondo NTC2018 [11.2.10.3]

    Per modulo elastico istantaneo del calcestruzzo va assunto quello secante tra
    la tensione nulla e 0,40fcm, determinato sulla base di apposite prove, da
    eseguirsi secondo la norma UNI EN 12390-13:2013.

    Il valore dovrà essere ridotto del 20% in caso di utilizzo di aggregati grossi
    di riciclo nei limiti previsti dalla Tab. 11.2.III. Tale formula non è
    applicabile ai calcestruzzi maturati a vapore.

    Essa non è da considerarsi vincolante nell’interpretazione dei controlli speri-
    mentali delle strutture.

    Unità in MPa

    Args:
        fck (float): Resistenza cilindrica caratteristica del calcestruzzo

    Returns:
        Ecm (float): Modulo elastico istantaneo del calcestruzzo
    """
    if type(fck) != float and type(fck) != int:
        raise ex.EXAExceptions("0001", "args fck must be float or int", type(fck))

    if fck < 0.0:
        raise ex.EXAExceptions("0002", "args fck must be > 0", fck)

    return 22000 * math.pow(fcm(fck) / 10, 0.3)


def rcSectionBeamAreaMax(Ac):
    """Area massima armature tese o compresse nelle travi secondo NTC2018 [4.1.6.11]

    Restituisce il valore dell'area massima di acciaio per una trave secondo
    NTC2018 che può essere disposta individualmente in zona tesa o compressa.

    Il valore è valido al di fuori delle zone di sovrapposizione per le quali è
    tollerabile il superamento.

    Unità in mm^2

    Args:
        Ac (float): Area di calcestruzzo. Ac > 0
    Returns:
        Asmax (float): Armatura minima
    """
    if type(Ac) != float:
        raise ex.EXAExceptions("0001", "args Ac must be float", type(Ac))
    if Ac < 0.0:
        raise ex.EXAExceptions("0002", "args Ac must be > 0", Ac)

    Asmax = 0.04 * Ac

    return Asmax


def rcSectionBeamAreaMin(
    fck: float,
    fyk: float,
    bt: float,
    d: float,
    crit: Literal["c1", "c2", "c12"] = "c12",
):
    """Area minima in zona tesa di armatura nelle travi secondo NTC2018 [4.1.6.11]

    Restituisce il valore dell'area minima di acciaio per una trave secondo
    NTC2018 nella zona tesa.

    Unità in N, mm

    Args:
        crit (Literal["c1","c2","c12"]): Criterio di minimo.
            Nel caso di:
            1. "c1" verrà usato 0.26 * fctm/fyk * bt * d
            2. "c2" verrà usato 0.0013 * bt * d
            3. "c12" verrà usato il massimo tra i due criteri
        d (float): Altezza utile della sezione
        bt (float): rappresenta la larghezza media della zona tesa;
            per una trave a T con piattabanda compressa, nel calcolare il
            valore di b t si considera solo la larghezza dell'anima;
        fyk (float): Snervamento dell'acciaio
        fck (float): Resistenza cilindrica caratteristica del calcestruzzo

    Returns:
        Asmin (float): Armatura minima
    """
    if type(fck) is not float and type(fck) is not int:
        raise ex.EXAExceptions("0001", "args fck must be float or int", type(fck))
    if fck < 0.0:
        raise ex.EXAExceptions("0002", "args fck must be > 0", fck)
    if type(fyk) is not float and type(fyk) is not int:
        raise ex.EXAExceptions("0003", "args fyk must be float or int", type(fyk))
    if fyk < 0.0:
        raise ex.EXAExceptions("0004", "args fyk must be > 0", fyk)
    if type(bt) is not float and type(bt) is not int:
        raise ex.EXAExceptions("0005", "args bt must be float or int", type(bt))
    if bt < 0.0:
        raise ex.EXAExceptions("0006", "args bt must be > 0", bt)
    if type(d) is not float and type(d) is not int:
        raise ex.EXAExceptions("0007", "args d must be float or int", type(d))
    if d < 0.0:
        raise ex.EXAExceptions("0008", "args d must be > 0", d)

    Asmin1 = 0.26 * fctm(fck) / fyk * bt * d
    Asmin2 = 0.0013 * bt * d

    if crit == "c1":
        return Asmin1
    if crit == "c2":
        return Asmin2
    if crit == "c12":
        return max(Asmin1, Asmin2)


def rebarArea(d: float):
    return d * d * math.pi / 4


def rebarNumber(As: float, d: float) -> Tuple[float, float]:
    areaRebar = rebarArea(d)
    nbRebar = math.ceil(As / areaRebar)
    return nbRebar, nbRebar * areaRebar / As


def ratioRebarWeight(Vs: float, Vc: float):
    return Vs * 7850 / Vc


def rcSectionBeamShearAreaMin(b):
    """Area minima staffe al metro nelle travi secondo NTC2018 [4.1.6.11]

    Restituisce il valore dell'area minima delle staffe per una trave secondo
    NTC2018

    Unità in mm

    Args:
        b (float): Larghezza minima dell'anima della trave in mm. b > 0
    Returns:
        Astmin (float): Armatura minima trasversale in mm2 al metro
    """
    if type(b) is not float and type(b) is not int:
        raise ex.EXAExceptions("0001", "args b must be float or int", type(b))
    if b <= 0.0:
        raise ex.EXAExceptions("0002", "args b must be > 0", b)

    Astmin = 1.5 * b

    return Astmin


def rcSectionBeamShearStepMax(d, dlmin):
    """Passo massimo delle staffe in mm nelle travi secondo NTC2018 [4.1.6.11]

    Restituisce il valore passo massimo delle staffe in mm nelle travi secondo
    NTC2018

    Unità in mm

    Args:
        d (float): Altezza utile della trave. d > 0
        dlmin (float): Diametro minimo delle barre longitudinali compresse.
            dlmin > 0
    Returns:
        smax (float): Passo minimo in mm
    """
    if type(d) != float:
        raise ex.EXAExceptions("0001", "args d must be float", type(d))
    if d <= 0.0:
        raise ex.EXAExceptions("0002", "args d must be > 0", d)
    if type(dlmin) != float:
        raise ex.EXAExceptions("0001", "args dlmin must be float", type(dlmin))
    if dlmin <= 0.0:
        raise ex.EXAExceptions("0002", "args dlmin must be > 0", dlmin)

    smax1 = 333.0
    smax2 = 0.8 * d
    # llp: vale solo per le compresse
    smax3 = 15 * dlmin

    smax = min(smax1, smax2, smax3)

    return smax


def rcSectionBeamShearStepsMax(d, dlmin):
    """Passo massimo delle staffe in mm nelle travi secondo NTC2018 [4.1.6.11]

    Restituisce il valore passo massimo delle staffe in mm nelle travi secondo
    NTC2018 come:
        1. minimo assoluto
        2. in rapporto all'altezza utile
        3. per le zone compresse

    Unità in mm

    Args:
        d (float): Altezza utile della trave. d > 0
        dlmin (float): Diametro minimo delle barre longitudinali compresse.
            dlmin > 0
    Returns:
        smax (float): Passo minimo in mm
    """
    if type(d) is not float and type(d) is not int:
        raise ex.EXAExceptions("0001", "args d must be float", type(d))
    if d <= 0.0:
        raise ex.EXAExceptions("0002", "args d must be > 0", d)
    if type(dlmin) is not float and type(dlmin) is not int:
        raise ex.EXAExceptions("0001", "args dlmin must be float", type(dlmin))
    if dlmin <= 0.0:
        raise ex.EXAExceptions("0002", "args dlmin must be > 0", dlmin)

    smax1 = 333.0
    smax2 = 0.8 * d
    # llp: vale solo per le compresse
    smax3 = 15 * dlmin

    return smax1, smax2, smax3


def rcSectionColumnAreaMin(fyk, Ac, Ned=0.0, gammas=1.15):
    """Area minima delle armature longitudinali nei pilastri NTC2018 [4.1.6.1.2]

    Restituisce il valore dell'area minima di acciaio per un pilasto secondo
    NTC2018.

    Unità in N, mm

    Args:
        fyk (float): Resistenza caratteristica dell'accciaio a snervamento > 0 [Mpa]
        Ac (float): Area del calcestruzzo della sezione compressa > 0 [mm^2]
        Ned (float): Sforzo normale di progetto della sezione > 0 [mm^2]
        gammas (float): Resistenza cilindrica caratteristica del calcestruzzo

    Returns:
        Asmin (float): Armatura minima
    """
    if type(fyk) != float:
        raise ex.EXAExceptions("0003", "args fyk must be float", type(fyk))
    if fyk < 0.0:
        raise ex.EXAExceptions("0004", "args fyk must be > 0", fyk)
    if type(Ac) != float:
        raise ex.EXAExceptions("0005", "args Ac must be float", type(Ac))
    if Ac < 0.0:
        raise ex.EXAExceptions("0006", "args Ac must be > 0", Ac)
    if type(Ned) != float:
        raise ex.EXAExceptions("0007", "args Ned must be float", type(Ned))
    if Ned < 0.0:
        raise ex.EXAExceptions("0008", "args Ned must Ned > 0", Ned)
    if type(gammas) != float:
        raise ex.EXAExceptions("0007", "args gammas must be float", type(gammas))
    if gammas < 0.0:
        raise ex.EXAExceptions("0008", "args gammas must gammas > 0", gammas)

    Asmin1 = 0.10 * Ned / (fyk / gammas)
    Asmin2 = 0.003 * Ac

    return max(Asmin1, Asmin2)


def rcSectionColumnAreaMax(Ac):
    """Area massima armature compresse nei pilastri secondo NTC2018 [4.1.6.1.2]

    Restituisce il valore dell'area massima di acciaio per un pilastro secondo
    NTC2018 che può essere disposta individualmente in zona tesa o su tutta l'area compressa.

    Il valore è valido al di fuori delle zone di sovrapposizione per le quali è
    tollerabile il superamento.

    Unità in mm^2

    Args:
        Ac (float): Area di calcestruzzo. Ac > 0
    Returns:
        Asmax (float): Armatura minima
    """
    if type(Ac) != float:
        raise ex.EXAExceptions("0001", "args Ac must be float", type(Ac))
    if Ac < 0.0:
        raise ex.EXAExceptions("0002", "args Ac must be > 0", Ac)

    Asmax = 0.04 * Ac

    return Asmax


def rcSectionColumnShearStepMax(dlmin):
    """Passo massimo delle staffe in mm nelle travi secondo NTC2018 [4.1.6.1.2]

    Restituisce il valore passo massimo delle staffe in mm nei pilastri secondo
    NTC2018

    Unità in mm

    Args:
        dlmin (float): Diametro minimo delle barre longitudinali.
            dlmin > 0
    Returns:
        smax (float): Passo minimo in mm
    """
    if type(dlmin) != float:
        raise ex.EXAExceptions("0001", "args dlmin must be float", type(dlmin))
    if dlmin <= 0.0:
        raise ex.EXAExceptions("0002", "args dlmin must be > 0", dlmin)

    smax1 = 250.0
    smax2 = 12.0 * dlmin

    smax = min(smax1, smax2)

    return smax


def rcSectionColumnShearDstMin(dlmax):
    """Diametro minimo staffe al metro nei pilastri secondo NTC2018 [4.1.6.1.2]

    Restituisce il valore dell'area minima delle staffe per una trave secondo
    NTC2018

    Unità in mm

    Args:
        dlmax (float): Diametro massimo armature longitudinali del pilastro. dlmax > 0
    Returns:
        dstmin (float): diametro minimo delle staffe
    """
    if type(dlmax) != float:
        raise ex.EXAExceptions("0001", "args dlmax must be float", type(dlmax))
    if dlmax <= 0.0:
        raise ex.EXAExceptions("0002", "args dlmax must be > 0", dlmax)

    dmin1 = 6.0
    dmin2 = dlmax / 4.0
    dstmin = max(dmin1, dmin2)

    return dstmin


def rcSectionColumnShearAreaMin(dlmin, dlmax):
    """Area minima staffe al metro nei pilastri secondo NTC2018 [4.1.6.1.2]

    Restituisce il valore dell'area minima delle staffe per una trave secondo
    NTC2018

    Unità in mm

    Args:
        dlmin (float): Diametro minimo armature longitudinali del pilastro. dlmin > 0
        dlmax (float): Diametro massimo armature longitudinali del pilastro. dlmax > 0
    Returns:
        Astmin (float): Armatura minima trasversale al metro
    """
    if type(dlmin) != float:
        raise ex.EXAExceptions("0001", "args dlmin must be float", type(dlmin))
    if dlmin <= 0.0:
        raise ex.EXAExceptions("0002", "args dlmin must be > 0", dlmin)
    if type(dlmax) != float:
        raise ex.EXAExceptions("0001", "args dlmax must be float", type(dlmax))
    if dlmax <= 0.0:
        raise ex.EXAExceptions("0002", "args dlmax must be > 0", dlmax)

    smax = rcSectionColumnShearStepMax(dlmin)
    dmin = rcSectionColumnShearDstMin(dlmax)

    Amin = 3.14 * dmin * dmin / 4.0
    Astmin = 2.0 * Amin / smax * 1000

    return Astmin


def shearCheckWithoutRebar(bw, d, fck, Asl, sigmacp=0.0, gammac=1.50, Ved=None, ll=1):
    """Verifica a taglio per elementi senza armatura specifica secondo NTC2018 [4.1.2.3.5.1]

    Se, sulla base del calcolo, non è richiesta armatura al taglio, è comunque
    necessario disporre un'armatura minima secondo quanto previsto al punto
    [4.1.6.1.1.] E' consentito omettere tale armatura minima in elementi quali
    solai, piastre e membrature a comportamento analogo, purché sia garantita
    una ripartizione trasversale dei carichi.

    Unità in N, mm

    Args:
        bw (float):
            Larghezza resistente a taglio > 0 [mm]. Minima della sezione.
        d (float):
            Altezza utile della sezione [mm].
        fck (float):
            Resistenza caratteristica cilindrica del cls > 0 [MPa]
        Asl (float):
            Armatura longitudinale in zona tesa il cui rapporto geometrico
            Asl/(bw*d )sarà <= 0.02
        sigmacp (:obj:`float`, optional):
            Tensione media di compressione della sezione [MPa] <= 0.2fcd.
            Se sigmacp < 0 lo sforzo è inteso di trazione. Default è 0.0
        gammac (:obj:`float`, optional):
            Coefficiente di sicurezza del calcestruzzo. Default è 1.50
        Ved (:obj:`float`, optional):
            Taglio sollecitante [N]. Default è None
        ll (:obj:`int`, optional):
            Log level. Default è 1

    Returns:
        res (dict): Risultato della verifica
            res['Vrd']   (float):
            res['check'] (float): Verifica se Vrd >= Vrd. 1 se vero 0 se falso
    """
    log("INF", "codeNTC2018-(shearCheckWithoutRebar) start ", ll)

    res = {"Vrd": 0, "check": None}

    if type(bw) != float and type(bw) != int:
        raise ex.EXAExceptions(
            "(shearCheckWithoutRebar)-0001", "args bw must be float or int", type(bw)
        )
    if type(d) != float and type(d) != int:
        raise ex.EXAExceptions(
            "(shearCheckWithoutRebar)-0001", "args d must be float or int", type(d)
        )
    if type(fck) != float and type(fck) != int:
        raise ex.EXAExceptions(
            "(shearCheckWithoutRebar)-0001", "args fck must be float or int", type(fck)
        )
    if type(Asl) != float and type(Asl) != int:
        raise ex.EXAExceptions(
            "(shearCheckWithoutRebar)-0001", "args Asl must be float or int", type(Asl)
        )
    if type(sigmacp) != float and type(sigmacp) != int:
        raise ex.EXAExceptions(
            "(shearCheckWithoutRebar)-0001",
            "args sigmacp must be float or int",
            type(sigmacp),
        )
    if type(gammac) != float and type(gammac) != int:
        raise ex.EXAExceptions(
            "(shearCheckWithoutRebar)-0001",
            "args gammac must be float or int",
            type(gammac),
        )
    if Ved is not None:
        if type(Ved) != float and type(Ved) != int:
            raise ex.EXAExceptions(
                "(shearCheckWithoutRebar)-0001",
                "args Ved must be float or int",
                type(Ved),
            )

    if bw < 0.0:
        raise ex.EXAExceptions(
            "(shearCheckWithoutRebar)-0001", "args bw must be > 0", bw
        )
    if d < 0.0:
        raise ex.EXAExceptions("(shearCheckWithoutRebar)-0001", "args d must be > 0", d)
    if fck < 0.0:
        raise ex.EXAExceptions(
            "(shearCheckWithoutRebar)-0001", "args fck must be > 0", fck
        )
    if Asl < 0.0:
        raise ex.EXAExceptions(
            "(shearCheckWithoutRebar)-0001", "args Asl must be > 0", Asl
        )
    if gammac <= 0.0:
        raise ex.EXAExceptions(
            "(shearCheckWithoutRebar)-0001", "args gammac must be > 0", gammac
        )
    if Ved is not None:
        if Ved < 0.0:
            raise ex.EXAExceptions(
                "(shearCheckWithoutRebar)-0001", "args Ved must be > 0", Ved
            )

    if sigmacp < 0:
        log("INF", "rol > 0.02 taked 0.02")

    k = 1 + math.sqrt(200 / d)
    if k > 2:
        k = 2
        log("INF", "k > 2 taked 2", ll)

    nimin = 0.035 * math.pow(k, 1.5) * math.pow(fck, 0.5)

    rol = Asl / (bw * d)
    if rol > 0.02:
        rol = 0.02
        log("INF", "rol > 0.02 taked 0.02")

    alphacc = 0.85
    fcd = alphacc * fck / gammac
    if sigmacp > 0.2 * fcd:
        sigmacp = 0.2 * fcd
        log("INF", "sigmacp > 0.2 * fcd taked 0.2 * fcd")

    Vrd1 = (
        (0.018 * k * math.pow(100 * rol * fck, 0.333) / gammac + 0.15 * sigmacp)
        * bw
        * d
    )
    Vrd2 = (nimin + 0.15 * sigmacp) * bw * d
    res["Vrd"] = max(Vrd1, Vrd2)

    if Ved is not None:
        if Ved < res["Vrd"]:
            res["check"] = 1
        else:
            res["check"] = 0

    log("INF", "codeNTC2018-(shearCheckWithoutRebar) end ", ll)

    return res


def shearCheckWithRebar(
    bw,
    d,
    fck,
    fyk,
    Asw,
    s,
    alpha=90.0,
    theta=None,
    sigmacp=0.0,
    gammac=1.50,
    gammas=1.15,
    Ved=None,
    ll=1,
) -> dict:
    """Verifica a taglio per elementi con armatura a taglio NTC2018 [4.1.2.3.5.2]

    La resistenza di progetto a taglio Vrd di elementi strutturali dotati di
    specifica armatura a taglio deve essere valutata sulla base di una adeguata
    schematizzazione a traliccio.
    Gli elementi resistenti dell'ideale traliccio sono: le armature trasversali,
    le armature longitudinali, il corrente compresso di calcestruzzo e i puntoni
    d'anima inclinati.

    Unità in N, mm

    Args:
        bw (float):
            Larghezza resistente a taglio > 0 [mm]. Minima della sezione.
        d (float):
            Altezza utile della sezione [mm].
        fck (float):
            Resistenza caratteristica cilindrica del cls > 0 [MPa]
        fyk (float):
            Resistenza caratteristica a snervamento dell'acciaio > 0 [MPa]
        Asw (float):
            Armatura trasversale e taglio disposta ad interasse s [mm^2]
        s (float):
              Passo dell'armatura trasversale [mm]
        alpha (:obj:`float`, optional):
            Angolo delle staffe o ferri longitudinali rispetto all'asse
            della trave [deg]. Default è 90.
        theta (:obj:`float`, optional):
            Angolo dei puntoni di calcestruzzo rispetto all'asse della
            trave [deg]. Default è None.
            Se theta = None allora viene calcolato l'angolo percui la
            resistenza del puntone eguaglia quella dell'acciaio se esiste
            una rottura bilanciata.
            Deve risultare 1 <= theta <= 2.5 [4.1.25] altrimenti viene preso il valore
            più vicino.
        sigmacp (:obj:`float`, optional):
            Tensione media di compressione della sezione [MPa] < fcd.
            Se sigmacp < 0 lo sforzo è inteso di trazione. Default è 0.0
        gammac (:obj:`float`, optional):
            Coefficiente di sicurezza del calcestruzzo. Default è 1.50
        gammas (:obj:`float`, optional):
            Coefficiente di sicurezza dell'acciaio. Default è 1.15
        Ved (:obj:`float`, optional):
            Taglio sollecitante [N]. Default è None
        ll (:obj:`int`, optional):
            Log level. Default è 1

    Returns:
        res (dict): Risultato della verifica
            res['Vrd'] (float):
            res['check'] (float): Verifica se Vrd >= Vrd. 1 se vero 0 se falso
            res['Vrsd'] (float): Resistenza del traliccio in acciaio [N]
            res['Vrcd'] (float): Resistenza del puntone in calcestruzzo [N]
            res['cotgTheta'] (float): Angolo di inclinazione dei puntoni
            res['al'] (float): Misura di traslazione del momento
    """
    log("INF", "codeNTC2018-(shearCheckWithRebar) start ", ll)

    if type(bw) not in (float, int):
        log("ERR", "shearCheckWithRebar(): args bw must be float", ll)
        raise ex.EXAExceptions("0001", "args bw must be float", type(bw))
    if bw < 0.0:
        log("ERR", "shearCheckWithRebar(): args bw must be > 0", ll)
        raise ex.EXAExceptions("0002", "args bw must be > 0", bw)

    if type(d) not in (float, int):
        log("ERR", "shearCheckWithRebar(): args d must be float", ll)
        raise ex.EXAExceptions("0001", "args d must be float", type(d))
    if d < 0.0:
        log("ERR", f"shearCheckWithRebar(): args d={d} must be > 0", ll)
        raise ex.EXAExceptions("0002", "args d must be > 0", d)

    if type(fck) not in (float, int):
        log("ERR", "args fck must be float", ll)
        raise ex.EXAExceptions("0001", "args fck must be float", type(fck))
    if fck < 0.0:
        log("ERR", "args fck must be > 0", ll)
        raise ex.EXAExceptions("0002", "args fck must be > 0", fck)

    if type(fyk) not in (float, int):
        log("ERR", "args fyk must be float", ll)
        raise ex.EXAExceptions("0001", "args fyk must be float", type(fyk))
    if fyk < 0.0:
        log("ERR", "args fyk must be > 0", ll)
        raise ex.EXAExceptions("0002", "args fyk must be > 0", fyk)

    if type(Asw) not in (float, int):
        log("ERR", "0001 - args Asw must be float", ll)
        raise ex.EXAExceptions("0001", "args Asw must be float", type(Asw))
    if Asw < 0.0:
        log("ERR", "0002 - args Asw must be > 0", ll)
        raise ex.EXAExceptions("0002", "args Asw must be > 0", Asw)

    if type(s) not in (float, int):
        log("ERR", "0003 - args s must be float", ll)
        raise ex.EXAExceptions("0003", "args s must be float", type(s))
    if s < 0.0:
        log("ERR", "0004 - args s must be > 0", ll)
        raise ex.EXAExceptions("0004", "args s must be > 0", s)

    if type(alpha) not in (float, int):
        log("ERR", f"args alpha is a {str(type(alpha)):s} must be float", ll)
        raise ex.EXAExceptions("0001", "args alpha must be float", type(alpha))
    if alpha < 45 or alpha > 90:
        log("ERR", f"args alpha must be 45 <= alpha = {alpha:.3f} <= 90", ll)
        raise ex.EXAExceptions("0002", "args alpha must be 45 <= alpha <= 90", s)

    if type(sigmacp) not in (float, int):
        log("ERR", "args sigmacp must be float", ll)
        raise ex.EXAExceptions("0001", "args sigmacp must be float", type(sigmacp))

    if type(gammac) not in (float, int):
        log("ERR", "args gammac must be float", ll)
        raise ex.EXAExceptions("0001", "args gammac must be float", type(gammac))
    if gammac < 0.0:
        log("ERR", "args gammac must be > 0", ll)
        raise ex.EXAExceptions("0002", "args gammac must be > 0", gammac)

    if type(gammas) not in (float, int):
        log("ERR", "args gammas must be float", ll)
        raise ex.EXAExceptions("0001", "args gammas must be float", type(gammas))

    if gammas < 0.0:
        log("ERR", "args gammas must be > 0", ll)
        raise ex.EXAExceptions("0002", "args gammas must be > 0", gammas)

    if Ved is not None:
        if type(Ved) not in (float, int):
            log("ERR", f"args Ved must be float. Ved is {type(Ved)}", ll)
            raise ex.EXAExceptions("0001", "args Ved must be float", type(Ved))
        if Ved < 0.0:
            log("ERR", "args Ved must be > 0", ll)
            raise ex.EXAExceptions("0002", "args Ved must be > 0", Ved)

    res = {
        "check": None,
        "Vrd": 0,
        "Vrsd": 0,
        "Vrcd": 0,
        "alpha": alpha,
        "cotgTheta": 0,
        "sigmacp": sigmacp,
        "al": 0,
        "alphac": 0,
        "err": False,
    }

    path = {"tetha_calculated": False}

    if theta is None:
        path["tetha_calculated"] = True
        cotgTheta = 1.11
    else:
        path["tetha_calculated"] = False
        cotgTheta = 1 / math.tan(math.radians(theta))
        if (cotgTheta < 1.0) or (cotgTheta > 2.5):
            log(
                "WRN",
                "cotgTheta = {:.3f}. It can not be less than 1 or greater then 2.5".format(
                    cotgTheta
                ),
                ll,
            )
            res["err"] = True
            return res

    cotgAlpha = 1 / math.tan(math.radians(alpha))
    alphar = math.radians(alpha)

    fyd = fyk / gammas
    fcd = fck / gammac

    res["path"] = path

    res["fyd"] = fyd
    res["fcd"] = fcd

    if sigmacp < 0:
        alphac = 1
    else:
        if (0 <= sigmacp) and (sigmacp < 0.25 * fcd):
            alphac = 1 + sigmacp / fcd
        elif (0.25 * fcd <= sigmacp) and (sigmacp <= 0.50 * fcd):
            alphac = 1.25
        elif (0.50 * fcd <= sigmacp) and (sigmacp < fcd):
            alphac = 2.5 * (1 - sigmacp / fcd)
        else:
            log(
                "WRN",
                "alphac can not be calulated cause sigmacp = {:.3f} > fcd = {:.3f}".format(
                    sigmacp, fcd
                ),
                ll,
            )
            res["err"] = True
            return res

    res["alphac"] = alphac

    log("INF", f"alphac = {alphac:.3f}", ll)

    # NTC2008 [4.1.2.3.5.2] ni si pone 0.5
    ni = 0.5

    def Vrsd_cotgTheta(x):
        return 0.9 * d * Asw / s * fyd * (cotgAlpha + x) * math.sin(math.radians(alpha))

    def Vrcd_cotgTheta(x):
        return 0.9 * d * bw * alphac * ni * fcd * (cotgAlpha + x) / (1 + math.pow(x, 2))

    res["bw"] = bw
    res["Asw"] = Asw
    res["s"] = s
    res["d"] = d

    if path["tetha_calculated"]:
        A = math.sin(alphar) * (Asw / s) * fyd
        B = bw * alphac * ni * fcd

        AA = A
        BB = A * cotgAlpha
        CC = A - B
        DD = (A - B) * cotgAlpha

        def T(x):
            return AA * pow(x, 3) + BB * pow(x, 2) + CC * pow(x, 1) + DD

        # First derivative of T(x)
        def D1_T(x):
            return 3 * AA * pow(x, 2) + 2 * BB * x + CC

        # D1_T(x) is ever monotone in [1.0,2.5] because vertex in x = -BB/(3*AA) is negative
        # than we are in growing branch of parabola
        if D1_T(1.0) * D1_T(2.5) > 0:
            log("INF", "Function T(x) is monotone ...", ll)
            if T(1.0) * T(2.5) < 0:
                log("INF", "Function T(x) has unique real solution ...", ll)
                path["method"] = "1. NUMPY with unique solution and T(x) monotone"

                # solving AA*x^3 + BB*x^2 + CC*x +DD
                roots = np.roots([AA, BB, CC, DD])
                log(
                    "INF",
                    "roots are x1 = {:.3f} x1 = {:.3f} x1 = {:.3f}".format(
                        roots[0], roots[1], roots[2]
                    ),
                    ll,
                )
                sol = []
                for r in roots:
                    if (1 <= r) and (r <= 2.5) and np.isreal(r):
                        sol.append(r)

                if len(sol) == 1:
                    cotgTheta = np.real(sol[0])
                    Vrd = Vrsd_cotgTheta(cotgTheta)
                    Vrsd = Vrsd_cotgTheta(cotgTheta)
                    Vrcd = Vrcd_cotgTheta(cotgTheta)
                else:
                    raise ex.EXAExceptions("0001", "Solution is not unique !!!", roots)

            else:
                log(
                    "INF",
                    "Function T(x) has not unique real solution but max and min value",
                    ll,
                )
                path[
                    "method"
                ] = "2. DISCRETE solution from 1.0 and 2.5 and T(x) monotone"
                cotgThetaLst = [1.0, 2.5]
                Vrd1 = min(Vrsd_cotgTheta(1.0), Vrcd_cotgTheta(1.0))
                Vrd2 = min(Vrsd_cotgTheta(2.5), Vrcd_cotgTheta(2.5))
                VrdLst = [Vrd1, Vrd2]
                Vrd = max(VrdLst)
                cotgTheta = cotgThetaLst[VrdLst.index(Vrd)]
                Vrsd = Vrsd_cotgTheta(cotgTheta)
                Vrcd = Vrcd_cotgTheta(cotgTheta)
        else:
            log("INF", "Function T(x) is not monotone ...", ll)
            log("INF", "... than we have max inside interval [1.0,2.5] ...", ll)
            path[
                "method"
            ] = "3. DISCRETE solution from 1.0 and 2.5 and T(x) not monotone"
            deltaOf_D1_T = 4 * pow(BB, 2) - 12 * AA * CC
            x1 = (-2 * BB - math.sqrt(deltaOf_D1_T)) / (2 * 3 * AA)
            x2 = (-2 * BB + math.sqrt(deltaOf_D1_T)) / (2 * 3 * AA)
            log(
                "INF",
                "Function D1_T(x) has zero in x1 = {:.3f} and x2 = {:.3f}".format(
                    x1, x2
                ),
                ll,
            )

            # Cause x1 is negative than x1 is outside [1.0,2.5] i use only x2
            if (1.0 <= x2) and (x2 <= 2.5):
                log("INF", f"x2 = {x2:.3f} inside [1.0,2.5] ...", ll)
            else:
                log("INF", f"x2 = {x2:.3f} outside [1.0,2.5] ...", ll)
                x2 = 1.0

            cotgThetaLst = [1.0, x2, 2.5]
            Vrd1 = min(Vrsd_cotgTheta(1.0), Vrcd_cotgTheta(1.0))
            Vrdx2 = min(Vrsd_cotgTheta(x2), Vrcd_cotgTheta(x2))
            Vrd2 = min(Vrsd_cotgTheta(2.5), Vrcd_cotgTheta(2.5))
            log(
                "INF",
                "... Vrd1 = {:.3f} Vrdx2 = {:.3f} Vrd2 = {:.3f}".format(
                    Vrd1, Vrdx2, Vrd2
                ),
                ll,
            )
            VrdLst = [Vrd1, Vrdx2, Vrd2]
            Vrd = max(VrdLst)
            cotgTheta = cotgThetaLst[VrdLst.index(Vrd)]
            Vrsd = Vrsd_cotgTheta(cotgTheta)
            Vrcd = Vrcd_cotgTheta(cotgTheta)

    else:
        Vrsd = Vrsd_cotgTheta(cotgTheta)
        Vrcd = Vrcd_cotgTheta(cotgTheta)
        Vrd = min(Vrsd, Vrcd)

    res["cotgTheta"] = cotgTheta
    res["Vrd"] = Vrd
    res["Vrsd"] = Vrsd
    res["Vrcd"] = Vrcd

    res["Ved"] = Ved
    if Ved is not None:
        if Ved < res["Vrd"]:
            res["check"] = 1
        else:
            res["check"] = 0

    # al is moment translation diagram (NTC 2018 4.1.30)
    res["al"] = 0.9 * d * res["cotgTheta"] / 2

    log("INF", "codeNTC2018-(shearCheckWithRebar) end ", ll)

    return res


class CrackOut:
    def __init__(self):
        self.__wk: float = 0.0
        self.__epsism: Union[None, float] = None
        self.__sigmas_stiffning: Union[None, float] = None
        self.__deltasm: Union[None, float] = None
        self.__deltasm1: Union[None, float] = None
        self.__deltasm2: Union[None, float] = None
        self.__roeff: Union[None, float] = None
        self.__zoneC: Union[None, float] = None
        self.__k2: Union[None, float] = None
        self.__alpham: Union[None, float] = None
        self.__epsi1: Union[None, float] = None
        self.__epsi2: Union[None, float] = None

    @property
    def epsi1(self):
        return self.__epsi1

    @epsi1.setter
    def epsi1(self, value):
        self.__epsi1 = value

    @property
    def epsi2(self):
        return self.__epsi2

    @epsi2.setter
    def epsi2(self, value):
        self.__epsi2 = value

    @property
    def alpham(self):
        return self.__alpham

    @alpham.setter
    def alpham(self, value):
        self.__alpham = value

    @property
    def k2(self):
        return self.__k2

    @k2.setter
    def k2(self, value):
        self.__k2 = value

    @property
    def zoneC(self):
        return self.__zoneC

    @zoneC.setter
    def zoneC(self, value):
        self.__zoneC = value

    @property
    def epsism(self):
        return self.__epsism

    @epsism.setter
    def epsism(self, value):
        self.__epsism = value

    @property
    def sigmas_stiffning(self):
        return self.__sigmas_stiffning

    @sigmas_stiffning.setter
    def sigmas_stiffning(self, value):
        self.__sigmas_stiffning = value

    @property
    def deltasm1(self):
        return self.__deltasm1

    @deltasm1.setter
    def deltasm1(self, value):
        self.__deltasm1 = value

    @property
    def deltasm(self):
        return self.__deltasm

    @deltasm.setter
    def deltasm(self, value):
        self.__deltasm = value

    @property
    def wk(self):
        return self.__wk

    @wk.setter
    def wk(self, value):
        self.__wk = value

    @property
    def deltasm2(self):
        return self.__deltasm2

    @deltasm2.setter
    def deltasm2(self, value):
        self.__deltasm2 = value

    @property
    def roeff(self):
        return self.__roeff

    @roeff.setter
    def roeff(self, value):
        self.__roeff = value

    def __str__(self):
        dispstr = f"__wk                = {self.__wk}\n"
        dispstr += f"__epsism            = {self.__epsism}\n"
        dispstr += f"__sigmas_stiffning  = {self.__sigmas_stiffning}\n"
        dispstr += f"__deltasm           = {self.__deltasm}\n"
        dispstr += f"__deltasm1          = {self.__deltasm1}\n"
        dispstr += f"__deltasm2          = {self.__deltasm2}\n"
        dispstr += f"__roeff             = {self.__roeff}\n"
        dispstr += f"__zoneC             = {self.__zoneC}\n"
        dispstr += f"__k2                = {self.__k2}\n"
        dispstr += f"__alpham            = {self.__alpham}\n"
        dispstr += f"__epsi1             = {self.__epsi1}\n"
        dispstr += f"__epsi2             = {self.__epsi2}\n"
        return dispstr

    def toDict(self) -> dict:
        return {
            "wk": self.__wk,
            "epsism": self.__epsism,
            "sigmas_stiffning": self.__sigmas_stiffning,
            "deltasm": self.__deltasm,
            "deltasm1": self.__deltasm1,
            "deltasm2": self.__deltasm2,
            "roeff": self.__roeff,
            "zoneC": self.__zoneC,
            "k2": self.__k2,
            "alpham": self.__alpham,
            "epsi1": self.__epsi1,
            "epsi2": self.__epsi2,
        }


def crackMeasure(
    epsiBot: float,
    epsiTop: float,
    deq: float,
    As: float,
    rebarsCover: float,
    rebarsDistance: float,
    hcEff: float,
    beff: float,
    hsec: float,
    xi: float,
    fck: float,
    sigmas: float,
    Es: float,
    load: str = "short",
) -> CrackOut:
    """Calcolo diretto ampiezza delle fessure secondo NTC2018 [C4.1.2.2.4.5]

    Unità in N, mm

    Args:
        deq (float): Diametro equivalente delle barre di armatura. Se ad esempio
            nella area efficace sono presenti due diametri d1 e d2 con rispetti
            vamente n1 ed n2 il numero di barre il valore proposto è:

                                    n1 * d1^2 + n2 * d2^2
                            deq = ---------------------
                                    n1 * d1 + n2 * d2

        As (float): Area totale dell'acciaio teso [mm^2]
        rebarsCover (float): Copriferro [mm]
        rebarsDistance (float): Interferro massimo [mm]
        hcEff (float): Altezza efficace [mm]
        beff (float): Larghezza efficace dell'area del tirante convenzionale [mm].
            Il valore è tale che moltiplicato per l'altezza heff fornisce l'area
            efficace del tirante convenzionale.
        hsec (float): Altezza della sezione [mm]
        xi (float): Profondità dell'asse neutro della sezione [mm]. Se x è positivo
            allora la sezione è parzializzata fino a quando x <= hsec. Se negativo
            la sezione è tutta tesa.
        fck (float): Resistenza cilindrica caratteristica a compressione del
            calcestruzzo in [MPa]
        sigmas (float): Tensione di trazione nell'acciaio [MPa]
        Es (float): Modulo di young dell'acciaio [MPa]
        epsiBot (float): Allungamento massimo sezione per sezione tutta tesa
        epsiTop (float): Allungamento minimo sezione per sezione tutta tesa
        load (str, optional): Durata del carico. 'short' o 'long'. Defaults to 'short'.

    Raises:
        ex.EXAExceptions: _description_

    Returns:
        dict: Risultato della misura epsism
            res['wk'] (float): Ampiezza caratteristica delle fessure [mm]
            res['epsism'] (float): Deformazione acciaio tra le fessure
            res['deltasm'] (float): Distanza media tra le fessure [mm]
            res['heff'] (float): Resistenza del puntone in calcestruzzo [N]
            res['roeff'] (float): Rapporto geometrico d'armatura efficace
            res['zoneC'] (float): Lunghezza per la quale è valida deltasm
    """
    res = CrackOut()

    if (epsiBot <= 0) and (epsiTop <= 0):
        res.wk = 0
        return res

    epsi1 = max(epsiBot, epsiTop)
    epsi2 = min(epsiBot, epsiTop) if min(epsiBot, epsiTop) > 0.0 else 0.0

    if load == "short":
        kt = 0.6
    elif load == "long":
        kt = 0.4
    else:
        raise ex.EXAExceptions(
            "0001", 'load parameter must be "short" or "long" !!!', load
        )

    Aceff = beff * hcEff
    roeff = As / Aceff
    alpham = Es / Ecm(fck)

    sigmas_stiffning = kt * fctm(fck) / roeff * (1 + alpham * roeff)
    if sigmas_stiffning >= 0.4 * sigmas:
        sigmas_stiffning = 0.4 * sigmas

    epsism = (sigmas - sigmas_stiffning) / Es

    # Medium distance between crack
    k1 = 0.8
    k2 = (epsi1 + epsi2) / (2 * epsi1)
    k3 = 3.4
    k4 = 0.425

    deltasm1 = (k3 * rebarsCover + k1 * k2 * k4 * deq / roeff) * (1 / 1.7)

    if epsiTop > epsiBot:
        deltasm2 = 0.75 * xi
    else:
        deltasm2 = 0.75 * (hsec - xi)

    zoneC = 5 * (rebarsCover + deq / 2)
    if rebarsDistance >= zoneC:
        deltasm = deltasm2
    else:
        deltasm = deltasm1

    wk = 1.7 * epsism * deltasm

    res.wk = wk
    res.epsism = epsism
    res.sigmas_stiffning = sigmas_stiffning
    res.deltasm = deltasm
    res.deltasm1 = deltasm1
    res.deltasm2 = deltasm2
    res.roeff = roeff
    res.zoneC = zoneC
    res.k2 = k2
    res.alpham = alpham
    res.epsi1 = epsi1
    res.epsi2 = epsi2
    return res


class WindActionZonesEnum(int, Enum):
    ZONE_1 = 1
    ZONE_2 = 2
    ZONE_3 = 3
    ZONE_4 = 4
    ZONE_5 = 5
    ZONE_6 = 6
    ZONE_7 = 7
    ZONE_8 = 8
    ZONE_9 = 9


class WindActionsZones(Table):
    descr_1 = (
        "Valle d'Aosta, Piemonte, Lombardia,"
        "Trentino Alto Adige, Veneto, "
        "Friuli Venezia Giulia "
        "(con l'eccezione della provincia di Trieste)"
    )
    descr_2 = "Emilia Romagna"
    descr_3 = (
        "Toscana, Marche, Umbria, Lazio, Abruzzo, "
        "Molise, Puglia, Campania, Basilicata, Calabria "
        "(esclusa la provincia di Reggio calabria)"
    )
    descr_4 = "Sicilia e provincia di Reggio Calabria"
    descr_5 = (
        "Sardegna (zona a oriente della retta congiungente"
        "Capo Teulada con l'Isola di Maddalena)"
    )
    descr_6 = (
        "Sardegna (zona a occidente della retta congiungente"
        "Capo Teulada con l'Isola di Maddalena)"
    )
    descr_7 = "Liguria"
    descr_8 = "Provincia di Trieste"
    descr_9 = "Isole (con l'eccezione di Sicilia e Sardegna) " "e mare aperto"
    tab01 = {
        "source": "Testo della Norma",
        "columns": ("id", "Zona", "Descrizione", "v_{b,0}", "a_{0}", "k_{a}"),
        "udm": ("", "", "", "m/s", "m", ""),
        "rows": (
            (1, "Zona 1", descr_1, 25, 1000, 0.40),
            (2, "Zona 2", descr_2, 25, 750, 0.45),
            (3, "Zona 3", descr_3, 27, 500, 0.37),
            (4, "Zona 4", descr_4, 28, 500, 0.36),
            (5, "Zona 5", descr_5, 28, 750, 0.40),
            (6, "Zona 6", descr_6, 28, 500, 0.36),
            (7, "Zona 7", descr_7, 28, 1000, 0.54),
            (8, "Zona 8", descr_8, 30, 1500, 0.50),
            (9, "Zona 9", descr_9, 31, 500, 0.32),
        ),
    }

    def __init__(self):
        Table.__init__(self, WindActionsZones.tab01)


class WindActionRoughtnessEnum(int, Enum):
    CLASS_A = 1
    CLASS_B = 2
    CLASS_C = 3
    CLASS_D = 4


class WindActionRoughtness(Table):
    descr_1 = (
        "Area urbana, in cui almeno il 15% della superficie "
        "del terreno sia coperto da edifici la cui altezza"
        "media supera i 15 m"
    )
    descr_2 = "Area urbana (non di classe A), suburbana, industriale o boschiva"
    descr_3 = (
        "Area con ostacoli diffusi (quali alberi, case, muri, recinzioni, ...); "
        "aree con rugosità non riconducibile alle Classi A, B, D."
    )
    descr_4 = (
        "a) Mare e relativa fascia costiera (entro 2 km dalla riva)"
        "b) Lago (con larghezza pari ad almeno 1 km) e relativa fascia costiera "
        "(entro 1 km dalla riva)"
        "c) Area priva di ostacoli o con al più rari ostacoli isolati (aperta "
        "campagna, aeroporti, aree agricole, pascoli, zone paludose o sabbiose, "
        "superfici innevate o ghiacciate, ...)"
    )
    tab01 = {
        "source": "Testo della Norma",
        "columns": ("id", "Classe di rugosità", "Descrizione"),
        "udm": ("", "", ""),
        "rows": (
            (1, "A", descr_1),
            (2, "B", descr_2),
            (3, "C", descr_3),
            (4, "D", descr_4),
        ),
    }

    def __init__(self):
        Table.__init__(self, WindActionRoughtness.tab01)


class WindActionSiteCategoriesEnum(int, Enum):
    CAT_I = 1
    CAT_II = 2
    CAT_III = 3
    CAT_IV = 4
    CAT_V = 5
    CAT_ND = 0


class WindActionSiteCategories(Table):
    tab01 = {
        "source": "Testo della Norma",
        "columns": ("id", "Categoria di esposizione", "k_r", "z_0", "z_min"),
        "udm": ("", "", "m", "m"),
        "rows": (
            (1, "I", 0.17, 0.01, 2),
            (2, "II", 0.19, 0.05, 4),
            (3, "III", 0.20, 0.10, 5),
            (4, "IV", 0.22, 0.30, 8),
            (5, "V", 0.23, 0.70, 12),
        ),
    }

    def __init__(self):
        Table.__init__(self, WindActionSiteCategories.tab01)


def windActionSiteCategorie(
    geoArea: WindActionZonesEnum,
    roughtness: WindActionRoughtnessEnum,
    aslm: float,
    seaDist: float,
) -> WindActionSiteCategoriesEnum:
    """Restituisce la categoria di sito [NTC2018 par. 3.3.1.]

    Args:
        geoArea (WindActionZonesEnum): zona geografica [NTC2018 par. 3.3.1.]
        roughtness (WindActionRoughtnessEnum): rugosità secondo [NTC2018 par. Tab.3.3.III]
        aslm (float): altezza sul livello del mare in m. Solo valori >= 0.
        seaDist (float): distanza dalla costa. Valori negativi
            significa in mare. Positivi nell'entroterra

    Returns:
        WindActionSiteCategoriesEnum: categoria del sito [NTC2018 fig. 3.3.2]
    """
    if aslm < 0:
        raise ex.EXAExceptions("ERR", "aslm must be > 0", aslm)

    c1 = any(
        [
            geoArea == WindActionZonesEnum.ZONE_1,
            geoArea == WindActionZonesEnum.ZONE_2,
            geoArea == WindActionZonesEnum.ZONE_3,
            geoArea == WindActionZonesEnum.ZONE_4,
            geoArea == WindActionZonesEnum.ZONE_5,
        ]
    )

    c2 = geoArea == WindActionZonesEnum.ZONE_6

    c3 = any(
        [geoArea == WindActionZonesEnum.ZONE_7, geoArea == WindActionZonesEnum.ZONE_8]
    )

    c4 = geoArea == WindActionZonesEnum.ZONE_9

    if c1:
        if seaDist <= 0:
            if roughtness != WindActionRoughtnessEnum.CLASS_D:
                raise ex.EXAExceptions(
                    "ERR",
                    "roughtness must be only D with 0 >= seaDist >= -2000",
                    seaDist,
                )

            if seaDist >= -2000:
                return WindActionSiteCategoriesEnum.CAT_I
            else:
                raise ex.EXAExceptions(
                    "ERR", "seaDist value must be >= -2000 m", seaDist
                )
        elif 0 < seaDist and seaDist <= 40000:
            if seaDist <= 10000:
                if roughtness == WindActionRoughtnessEnum.CLASS_A:
                    return WindActionSiteCategoriesEnum.CAT_IV
                elif roughtness == WindActionRoughtnessEnum.CLASS_B:
                    return WindActionSiteCategoriesEnum.CAT_III
                elif roughtness == WindActionRoughtnessEnum.CLASS_C:
                    if geoArea == WindActionZonesEnum.ZONE_5:
                        return WindActionSiteCategoriesEnum.CAT_III
                    else:
                        return WindActionSiteCategoriesEnum.CAT_II
                else:
                    return WindActionSiteCategoriesEnum.CAT_II

            else:
                if roughtness == WindActionRoughtnessEnum.CLASS_A:
                    return WindActionSiteCategoriesEnum.CAT_IV
                elif roughtness == WindActionRoughtnessEnum.CLASS_B:
                    return WindActionSiteCategoriesEnum.CAT_III
                elif roughtness == WindActionRoughtnessEnum.CLASS_C:
                    return WindActionSiteCategoriesEnum.CAT_III
                else:
                    return WindActionSiteCategoriesEnum.CAT_II
        else:
            if aslm < 500:
                if roughtness == WindActionRoughtnessEnum.CLASS_A:
                    return WindActionSiteCategoriesEnum.CAT_V
                elif roughtness == WindActionRoughtnessEnum.CLASS_B:
                    return WindActionSiteCategoriesEnum.CAT_IV
                elif roughtness == WindActionRoughtnessEnum.CLASS_C:
                    return WindActionSiteCategoriesEnum.CAT_III
                else:
                    return WindActionSiteCategoriesEnum.CAT_II
            elif aslm < 750:
                if roughtness == WindActionRoughtnessEnum.CLASS_A:
                    return WindActionSiteCategoriesEnum.CAT_V
                elif roughtness == WindActionRoughtnessEnum.CLASS_B:
                    return WindActionSiteCategoriesEnum.CAT_IV
                elif roughtness == WindActionRoughtnessEnum.CLASS_C:
                    return WindActionSiteCategoriesEnum.CAT_IV
                else:
                    return WindActionSiteCategoriesEnum.CAT_III
            else:
                if roughtness == WindActionRoughtnessEnum.CLASS_A:
                    return WindActionSiteCategoriesEnum.CAT_IV
                elif roughtness == WindActionRoughtnessEnum.CLASS_B:
                    return WindActionSiteCategoriesEnum.CAT_IV
                elif roughtness == WindActionRoughtnessEnum.CLASS_C:
                    if geoArea == WindActionZonesEnum.ZONE_1:
                        return WindActionSiteCategoriesEnum.CAT_IV
                    else:
                        return WindActionSiteCategoriesEnum.CAT_III

    elif c2:
        if seaDist <= 0:
            if roughtness != WindActionRoughtnessEnum.CLASS_D:
                raise ex.EXAExceptions(
                    "ERR",
                    "roughtness must be only D with 0 >= seaDist >= -2000",
                    seaDist,
                )

            if seaDist >= -2000:
                return WindActionSiteCategoriesEnum.CAT_I
            else:
                raise ex.EXAExceptions(
                    "ERR", "seaDist value must be >= -2000 m", seaDist
                )

        elif 0 < seaDist and seaDist <= 40000:
            if seaDist <= 10000:
                if roughtness == WindActionRoughtnessEnum.CLASS_A:
                    return WindActionSiteCategoriesEnum.CAT_III
                elif roughtness == WindActionRoughtnessEnum.CLASS_B:
                    return WindActionSiteCategoriesEnum.CAT_II
                elif roughtness == WindActionRoughtnessEnum.CLASS_C:
                    return WindActionSiteCategoriesEnum.CAT_II
                else:
                    return WindActionSiteCategoriesEnum.CAT_I
            else:
                if roughtness == WindActionRoughtnessEnum.CLASS_A:
                    return WindActionSiteCategoriesEnum.CAT_IV
                elif roughtness == WindActionRoughtnessEnum.CLASS_B:
                    return WindActionSiteCategoriesEnum.CAT_III
                elif roughtness == WindActionRoughtnessEnum.CLASS_C:
                    return WindActionSiteCategoriesEnum.CAT_III
                else:
                    return WindActionSiteCategoriesEnum.CAT_II
        else:
            if aslm < 500:
                if roughtness == WindActionRoughtnessEnum.CLASS_A:
                    return WindActionSiteCategoriesEnum.CAT_V
                elif roughtness == WindActionRoughtnessEnum.CLASS_B:
                    return WindActionSiteCategoriesEnum.CAT_IV
                elif roughtness == WindActionRoughtnessEnum.CLASS_C:
                    return WindActionSiteCategoriesEnum.CAT_III
                else:
                    return WindActionSiteCategoriesEnum.CAT_II
            else:
                if roughtness == WindActionRoughtnessEnum.CLASS_A:
                    return WindActionSiteCategoriesEnum.CAT_V
                elif roughtness == WindActionRoughtnessEnum.CLASS_B:
                    return WindActionSiteCategoriesEnum.CAT_IV
                elif roughtness == WindActionRoughtnessEnum.CLASS_C:
                    return WindActionSiteCategoriesEnum.CAT_IV
                else:
                    return WindActionSiteCategoriesEnum.CAT_III

    elif c3:

        if seaDist <= 0:
            if roughtness != WindActionRoughtnessEnum.CLASS_D:
                raise ex.EXAExceptions(
                    "ERR",
                    "roughtness must be only D with 0 >= seaDist >= -2000",
                    seaDist,
                )

            if seaDist > -500:
                return WindActionSiteCategoriesEnum.CAT_I
            elif seaDist >= -2000:
                return WindActionSiteCategoriesEnum.CAT_II
            else:
                raise ex.EXAExceptions(
                    "ERR", "seaDist value must be >= -2000 m", seaDist
                )
        else:
            if roughtness == WindActionRoughtnessEnum.CLASS_A:
                return WindActionSiteCategoriesEnum.CAT_IV
            elif roughtness == WindActionRoughtnessEnum.CLASS_B:
                return WindActionSiteCategoriesEnum.CAT_IV
            elif roughtness == WindActionRoughtnessEnum.CLASS_C:
                return WindActionSiteCategoriesEnum.CAT_III
            else:
                if geoArea == WindActionZonesEnum.ZONE_8:
                    return WindActionSiteCategoriesEnum.CAT_II
                else:
                    return WindActionSiteCategoriesEnum.CAT_III

    elif c4:
        if seaDist <= 0:
            if roughtness != WindActionRoughtnessEnum.CLASS_D:
                raise ex.EXAExceptions(
                    "ERR",
                    "roughtness must be only D with 0 >= seaDist >= -2000",
                    seaDist,
                )
            else:
                return WindActionSiteCategoriesEnum.CAT_I
        else:
            return WindActionSiteCategoriesEnum.CAT_I

    else:
        raise ex.EXAExceptions("ERR", "area unknown !!!")

    return WindActionSiteCategoriesEnum.CAT_ND


class WindActionsExpositionParam(Table):
    tab01 = {
        "source": "Testo della Norma",
        "columns": ("id", "Categoria di esposizione", "K_r", "z_0", "z_{min}"),
        "udm": ("", "", "", "m", "m"),
        "rows": (
            (1, "I", 0.17, 0.01, 2),
            (2, "II", 0.19, 0.05, 4),
            (3, "III", 0.20, 0.10, 5),
            (4, "IV", 0.22, 0.30, 8),
            (5, "V", 0.23, 0.70, 12),
        ),
    }

    def __init__(self):
        Table.__init__(self, WindActionsExpositionParam.tab01)


def windActionsExposition(
    siteCategory: WindActionSiteCategoriesEnum, z: float
) -> float:
    """Restituisce il coefficienti di esposizione [NTC20018 3.3.7.]

    Il valore del coefficiente di topografia è assunto unitario.

    Args:
        siteCategory (WindActionSiteCategoriesEnum): categoria di esposizione del sito
        z (float): altezza della costruzione dal suolo. z > 0 && z < 200

    Raises:
        ex.EXAExceptions: z must be > 0
        ex.EXAExceptions: z must be < 200

    Returns:
        float: coefficiente di esposizione
    """
    if z < 0:
        raise ex.EXAExceptions("ERR", "z must be > 0", z)

    if z > 200:
        raise ex.EXAExceptions("ERR", "z must be < 200", z)

    def ce(Kr: float, z0: float, zmin: float, z: float) -> float:
        ct = 1.0

        if z < zmin:
            z = zmin

        return Kr * Kr * ct * np.log(z / z0) * (7 + ct * np.log(z / z0))

    param = WindActionsExpositionParam()
    Kr = param.valueAt(siteCategory.value)[2]
    z0 = param.valueAt(siteCategory.value)[3]
    zmin = param.valueAt(siteCategory.value)[4]

    return ce(Kr, z0, zmin, z)


def windActionsPressure(
    zone: WindActionZonesEnum,
    roughtness: WindActionRoughtnessEnum,
    aslm: float,
    seaDist: float,
    zRef: float,
    Tr: float = 50,
) -> dict:
    """Calcola la pressione cinetica di picco e tutte le grandezze correlate

    Args:
        zone (WindActionZonesEnum): _description_
        roughtness (WindActionRoughtnessEnum): _description_
        aslm (float): _description_
        seaDist (float): _description_
        zRef (float): è l'altezza di riferimento in metri del filetto fluido
            considerato.
        Tr (float, optional): _description_. Defaults to 50.

    Returns:
        dict: _description_
    """
    res = {}

    zoneTable = WindActionsZones()
    zone_label = zoneTable.valueAt(zone)[1]
    zone_descr = zoneTable.valueAt(zone)[2]
    zone_vb0 = zoneTable.valueAt(zone)[3]
    zone_a0 = zoneTable.valueAt(zone)[4]
    zone_ka = zoneTable.valueAt(zone)[5]

    rougthnessTable = WindActionRoughtness()
    rougthness_label = rougthnessTable.valueAt(roughtness)[1]
    rougthness_descr = rougthnessTable.valueAt(roughtness)[2]

    res["input"] = {
        "zone_label": zone_label,
        "zone_descr": zone_descr,
        "zone_vb0": zone_vb0,
        "zone_a0": zone_a0,
        "zone_ka": zone_ka,
        "rougthness_label": rougthness_label,
        "rougthness_descr": rougthness_descr,
        "aslm": aslm,
        "seaDist": seaDist,
        "zRef": zRef,
        "Tr": Tr,
    }

    # (1) calcolo velocità base di riferimento
    zoneParam = WindActionsZones()
    zoneLabel = zoneParam.valueAt(zone.value)[1]
    zoneDescr = zoneParam.valueAt(zone.value)[2]
    vb0 = zoneParam.valueAt(zone.value)[3]
    a0 = zoneParam.valueAt(zone.value)[4]
    ka = zoneParam.valueAt(zone.value)[5]

    # (1.1) coefficiente di altitudine
    if aslm <= a0:
        ca = 1
    else:
        ca = 1 + ka * (aslm / a0 - 1)

    # (1.2) velocità di riferimento
    vb = vb0 * ca

    # (2) calcolo velocità di riferimento di progetto

    # (2.1) coefficiente del tempo di ritorno
    cr = 0.75 * np.sqrt(1 - 0.2 * np.log(-np.log(1 - 1 / Tr)))

    # (2.2) velocità di riferimento
    vr = vb * cr

    # (3) pressione cinetica di picco

    # (3.1) calcolo coefficiente di esposizione
    siteCategorie = windActionSiteCategorie(zone, roughtness, aslm, seaDist)
    ce = windActionsExposition(siteCategorie, zRef)

    # (3.2) densità dell'aria in kg/mc
    rho = 1.25

    # (3.3) pressione di riferimento
    qr = 0.5 * vr * vr * rho / 1000

    # (3.3) pressione cinetica di picco
    #
    # La pressione cinetica di picco del vento q p è il valore atteso della
    # pressione cinetica massima del vento sull’intervallo di tempo T = 10
    # minuti. Essa dipende dall’altezza z sul suolo, dalla ventosità della
    # zona in esame (paragrafo 3.2.1), dal periodo di ritorno di progetto
    # (paragrafo 3.2.2), dalle caratteristiche locali del sito ove sorge
    # la costruzione e dalla densità dell’aria.
    #
    # Dalla formula le udm sono N/mq ma noi trasformiamo in KN/mq
    qp = qr * ce

    siteCat = WindActionSiteCategories()
    siteCatLabel = siteCat.valueAt(siteCategorie.value)[1]
    k_r = siteCat.valueAt(siteCategorie.value)[2]
    z_0 = siteCat.valueAt(siteCategorie.value)[3]
    z_min = siteCat.valueAt(siteCategorie.value)[4]

    res["output"] = {
        "zone_label": zoneLabel,
        "zone_descr": zoneDescr,
        "zone_vb0": vb0,
        "zone_a0": a0,
        "zone_ka": ka,
        "ca": ca,
        "vb": vb,
        "cr": cr,
        "vr": vr,
        "siteCat_label": siteCatLabel,
        "siteCat_k_r": k_r,
        "siteCat_z_0": z_0,
        "siteCat_z_min": z_min,
        "ce": ce,
        "qr": qr,
        "rho": rho,
        "qp": qp,
    }

    return res


def windActionsForcesOnRect(d_shape: float, h_shape: float, qp: float = 1.0) -> dict:
    """Azione del vento su elementi orizzontali rettangolari [CNR DT207/2008 G.10.3]

    Fornisce l'azione del vento su elementi orizzontali rettangolari secondo la
    direzione X (Y dal basso verso l'alto).

                        ^     |------------------------|
        vento X         |     |                        |
        ------>      h_shape  |                        |
                        |     |                        |
                        v     |------------------------|

                              |<------- d_shape ------>|

    Args:
        d_shape (float): Larghezza del rettangolo
        h_shape (float): Altezza del rettangolo
        qp (float, optional): Pressione cinetica di picco. Defaults to 1.0

    Returns:
        dict:
            'input':
                'd_shape': larghezza elemento
                'h_shape': altezza elemento,
                'qp': pressione cinetica di picco
            'output':
                'delta': rapporto d_shape/h_shape,
                'lref': lunghezza di riferimento,
                'C_fx': coefficiente di forza lungo X,
                'C_fy': coefficiente di forza lungo Y,
                'C_mz': coefficiente di momento lungo Z,
                'f_x': forza lungo X per unità di lunghezza,
                'f_y': forza lungo Y per unità di lunghezza,
                'm_z': coppia lungo Z per unità di lunghezza,
    """
    delta = d_shape / h_shape

    if 0.1 <= delta and delta < 0.2:
        C_fx0 = 2.0

    elif 0.2 <= delta and delta < 0.7:
        C_fx0 = 0.73 * np.log10(delta) + 2.51

    elif 0.7 <= delta and delta < 5.0:
        C_fx0 = -1.64 * np.log10(delta) + 2.15

    elif 5.0 <= delta and delta < 10.0:
        C_fx0 = -0.33 * np.log10(delta) + 1.23

    elif 10.0 <= delta and delta <= 50.0:
        C_fx0 = 0.9

    else:
        raise ex.EXAExceptions("ERR", "must be 0.1 <= d_shape/h_shape <= 50.0", delta)

    C_fy0 = 0.0
    C_mz0 = 0.0

    # A favore di sicurezza poichè compreso tra 0 ed 1 [CNR DT207/2008 G.9.4]
    g_lambda = 1.0

    # Coefficienti di forza
    C_fx = C_fx0 * g_lambda
    C_fy = C_fy0 * g_lambda
    C_mz = C_mz0 * g_lambda

    # Lunghezza di riferimento
    lref = h_shape

    f_x = qp * lref * C_fx
    f_y = qp * lref * C_fy
    m_z = qp * lref * lref * C_mz

    res = {
        "input": {"d_shape": d_shape, "h_shape": h_shape, "qp": qp},
        "output": {
            "delta": delta,
            "lref": lref,
            "C_fx": C_fx,
            "C_fy": C_fy,
            "C_mz": C_mz,
            "f_x": f_x,
            "f_y": f_y,
            "m_z": m_z,
        },
    }

    return res


def windActionsForcesOnBridge(d_bridge: float, h_tot: float, qp: float = 1) -> dict:
    """Azione del vento su impalcati da ponte [CNR DT207/2008 G.11]

    Calcola l'azione del vento su impalcati da ponte secondo la CNR DT207/2008
    al punto G.11.

    Sono applicabili a impalcati da ponte a sezione costante lungo la linea
    d'asse, limitatamente ai tipi riportati in Figura G.54, per ponti a luce
    singola o multipla, purché di lunghezza non superiore a 200 m.

    Si considerano esclusi da questo paragrafo altri tipi di ponti quali i
    ponti ad arco, i ponti sospesi o strallati, i ponti coperti o mobili,
    i ponti con curvature planimetriche significative e i ponti costituiti
    da più impalcati affiancati non riconducibili al semplice schema illustrato
    nel paragrafo G.11.2. Per tutti questi tipi strutturali occorre sviluppare
    specifiche e mirate valutazioni.

    Il calcolo è valido per valori di d/h >=2. Per valori inferiori si fà
    riferimento a windActionsForcesOnRect() per C_fx

    Args:
        d_bridge (float): _description_
        h_tot (float): _description_
        qp (float, optional): _description_. Defaults to 1.

    Returns:
        dict: _description_
    """
    delta = d_bridge / h_tot

    if delta < 2.0:
        resRect = windActionsForcesOnRect(d_bridge, h_tot, qp)
        C_fx = resRect["output"]["C_fx"]
        C_fy = 0.70 + 0.1 * delta

    elif 2 <= delta and delta <= 5.0:
        C_fx = 1.85 / delta - 0.10
        C_fy = 0.70 + 0.1 * delta

    else:
        C_fx = 1.35 / delta
        C_fy = 1.20

    C_mz = 0.20

    # Lunghezza di riferimento
    lref = d_bridge

    f_x = qp * lref * C_fx
    f_y = qp * lref * C_fy
    m_z = qp * lref * lref * C_mz

    res = {
        "input": {"d_bridge": d_bridge, "h_tot": h_tot, "qp": qp},
        "output": {
            "delta": delta,
            "lref": lref,
            "C_fx": C_fx,
            "C_fy": C_fy,
            "C_mz": C_mz,
            "f_x": f_x,
            "f_y": f_y,
            "m_z": m_z,
        },
    }

    return res



from typing import List, Union


class Fun:
    def __init__(
        self, left: Union[float, None] = None, right: Union[float, None] = None
    ):
        self.__left = left
        self.__right = right

    @property
    def left(self):
        return self.__left

    @left.setter
    def left(self, val):
        if isinstance(val, float) or val is None:
            if self.__left is not None:
                if self.__right is not None and self.__left <= self.__right:
                    self.__left = val
                elif self.__right is None:
                    self.__left = val
                else:
                    raise ValueError("Left must be less or equal then Right")
            else:
                self.__left = val
        else:
            raise ValueError("Left must be float or None type")

    @left.deleter
    def left(self):
        del self.__left

    @property
    def right(self):
        return self.__right

    @right.setter
    def right(self, val):
        if isinstance(val, float) or val is None:
            if self.__right is not None:
                if self.__left is not None and not self.__left > self.__right:
                    self.__right = val
                elif self.__left is None:
                    self.__right = val
                else:
                    print(f"Left {self.__left}")
                    print(f"Right {self.__right}")
                    raise ValueError("Right must be greater or equal then Right")
            else:
                self.__right = val
        else:
            raise ValueError("Right must be float or None type")

    @right.deleter
    def right(self):
        del self.__right

    def eval(self, x: float) -> float:
        raise NotImplementedError("eval() not implemented for subclass !!!")


class piecewiseFun(Fun):
    def __init__(
        self, x: Union[List[float], None] = None, fx: Union[List[float], None] = None
    ):
        super().__init__()
        if x is None:
            x = []

        if fx is None:
            fx = []

        self.__x = x
        self.__fx = fx

        if len(x) != len(fx):
            raise ValueError(f"Len of x <{self.__x}> must be same of fx <{self.__fx}>")

        if len(x) > 0:
            self.left = self.__x[0]
            self.right = self.__x[len(self.__x) - 1]

    @property
    def x(self):
        return self.__x

    @x.setter
    def x(self, val):
        if isinstance(val, list):
            for v in val:
                if not isinstance(v, float):
                    raise ValueError(f"The value {v} must be float")
            self.__x = val
        else:
            raise ValueError("Need to assign only list of float to piecewiseFun")

    @x.deleter
    def x(self):
        del self.__x

    @property
    def fx(self):
        return self.__fx

    @fx.setter
    def fx(self, val):
        if isinstance(val, list):
            for v in val:
                if not isinstance(v, float):
                    raise ValueError(f"The value {v} must be float")
            self.__fx = val
        else:
            raise ValueError("Need to assign only list of float to piecewiseFun")

    @fx.deleter
    def fx(self):
        del self.__fx

    def eval(self, x: float) -> float:

        if len(self.__x) != len(self.__fx):
            raise ValueError(f"Len of x <{self.__x}> must be same of fx <{self.__fx}>")

        self.left = self.__x[0]
        self.right = self.__x[len(self.__x) - 1]

        for i in range(0, len(self.__x) - 1):
            x1 = self.__x[i]
            x2 = self.__x[i + 1]
            fx1 = self.__fx[i]
            fx2 = self.__fx[i + 1]
            if x2 == x1:
                raise ValueError(f"Value x2 <{x1}> is same of x1")
            if x > x1 and x < x2:
                A = fx2 - fx1 / (x2 - x1)
                B = fx1 - A * x1
                return A * x + B
            elif x == x1:
                return fx1
            elif x == x2:
                return fx2
            else:
                pass

        raise ValueError(
            "Value x must inside ({},{})".format(
                self.__x[0], self.__x[len(self.__x) - 1]
            )
        )



import sys
import pytest
from pathlib import Path

from pycivil.EXAGeotechnical import base as geot
from pycivil.EXAParametric import box as param

def test_001(tmp_path: Path) -> None:

    E_top = geot.ModulusYoung()
    E_top.setChoose(9)
    Et = (E_top.value()[2] + E_top.value()[2]) / 2 * 0.1

    E_bot = geot.ModulusYoung()
    E_bot.setChoose(2)
    Eb = (E_bot.value()[2] + E_bot.value()[2]) / 2 * 0.1

    ni_top = geot.RatioPoisson()
    ni_top.setChoose(5)
    nit = (ni_top.value()[2] + ni_top.value()[2]) / 2

    ni_bot = geot.RatioPoisson()
    ni_bot.setChoose(3)
    nib = (ni_bot.value()[2] + ni_bot.value()[2]) / 2

    soilTop = geot.SoilLayer("Terreno superiore - Ghiaia")
    soilTop.setEt(Et)
    soilTop.setNit(nit)
    soilTop.setDry(gamma=20, phi=30, coe=0)
    print(soilTop)

    soilBot = geot.SoilLayer("Terreno inferiore - Argilla Limosa")
    soilBot.setEt(Eb)
    soilBot.setNit(nib)
    soilBot.setCu(0.05)
    print(soilBot)

    title = "CENTROPADANA - tombino sottostrada - manufatto I.15"
    box = param.BoxTubeShape01(
        descr=title, B=2500 + 300, H=2000 + 300, Tw=300, Tt=300, Tb=300
    )

    box.assignTopLayer(soilTop)
    box.assignBotLayer(soilBot)

    box.springs()
    print(box)

    box.assignSeismicAction(0.261, 0.2)

    box.setTopCover(5000.0 - 600.0 + 150.0)
    box.setPaverThickness(600.0, 20.0)

    box.buildFEModel()

    fileName = "I15"
    fileExtension = ".mgt"
    box.exportFEModel(str(tmp_path / f"{fileName}"), fileExtension)
    filePath = tmp_path / f"{fileName}{fileExtension}"
    box.model().clear()

    assert filePath.exists()


if __name__ == "__main__":
    sys.exit(pytest.main(Path(__file__)))




from typing import Literal, List, Any, Union, Dict
from pathlib import Path
import pandas as pd
import numpy as np
import math
import time
import json

from numpy.typing import ArrayLike, NDArray

import pyvista as pv
from pyvista import Plotter

import matplotlib.pyplot as plt

from pycivil.EXAUtils.logging import log
from pycivil.EXAStructural.templateRCRect import RCTemplRectEC2
from pycivil.EXAStructural.codes import Code
from pycivil.EXAStructural.loads import Frequency_Enum, LimiteState_Enum, ForcesOnSection as Load
from pycivil.EXAStructural.materials import Concrete, ConcreteSteel
from pycivil.EXAStructuralCheckable.RcsRectangular import RcsRectangular
from pycivil.EXAStructural.checkable import CheckableCriteriaEnum as Crit, Enum

class ResultsType(str, Enum):
    WA = "WA",
    FORCE = "FORCE"

class PostProcessor:
    def __init__(self):

        self.__ll: Literal[0, 1, 2, 3] = 3

        self.__working_path: None | Path = None
        self.__in_rel_path: None | Path = Path("in")
        self.__out_rel_path: None | Path = Path("out")

        self.__fn_nodes: str = "Nodes_Coordinates.txt"
        self.__fn_elements: str = "Elements_Elements.txt"
        self.__fn_model: str = "Model.txt"

        self.__fn_assignments_for_rebar: str = "rebars_assignment.xlsx"
        self.__fn_assignments_for_check: str = "check_assignment.xlsx"

        self.__nodes_df: None | pd.DataFrame = None
        self.__nodes_id: None | np.ndarray = None
        self.__nodes_coords: None | np.ndarray = None

        self.__elements_df: None | pd.DataFrame = None
        self.__elements_2D_id: None | np.ndarray = None
        self.__elements_2D_groups: None | np.ndarray = None
        self.__elements_2D_properties: None | dict = None
        self.__elements_2D_connectivity: None | List[np.ndarray] = None
        self.__elements_2D_type: None | List[np.ndarray] = None

        self.__assignments_for_rebar: None | dict = None
        self.__assignments_for_check: None | dict = None

        self.__model_static_nl_combo: None | List[str] = None
        self.__model_plate_properties: None | dict = None

        self.__plate_results_stream: None | List[Any] = None
        self.__plate_results_log_stream: None | List[dict] = None
        self.__idx_check_nook: None | List[int] = None

        self.__results_parsed: Union[None, NDArray] = None

        self.__filter_on_group: List[str] | None = None

        self.__curr_check_required: Crit = Crit.SLU_NM
        self.__curr_input_results_tp: ResultsType = ResultsType.WA
        self.__curr_input_results_name: str = ""

        # Main results dictionary
        self.__results_checked: Dict[str, Dict[ResultsType, Dict[Crit, List[Any]]]] = dict()

        self.__path_last_figure_plotted: Path | None = None

    def setFilterOnGroups(self, f: List[str] | None):
        self.__filter_on_group = f

    def getFilterOnGroups(self) -> List[str] | None:
        return self.__filter_on_group

    def getResultsChecked(self) -> Dict[str, Dict[ResultsType, Dict[Crit, List[Any]]]]:
        return self.__results_checked

    def getPathLastFigurePlotted(self) -> Path | None:
        return self.__path_last_figure_plotted

    def getPathOutAbsolute(self) -> Path | None:
        return self.__working_path / self.__out_rel_path

    def getPathWorking(self) -> Path | None:
        return self.__working_path


    @staticmethod
    def __parse_model_plate_properties(path: Path) -> None | dict:
        parsed = {}
        read_block = False
        with open(path) as file:
            for item in file:
                if not read_block:
                    if item[0] == '/':
                        if "/ PLATE PROPERTIES" in item:
                            read_block = True
                else:
                    if "PlateShellProp" in item:
                        row_split = item.split()
                        prop_id = int(row_split[1])
                        prop_name = row_split[3]
                        parsed[prop_id] = []
                        parsed[prop_id].append(prop_name)
                    if " MemThick" in item:
                        row_split = item.split()
                        prop_MemThick = float(row_split[1])
                        parsed[prop_id].append(prop_MemThick)
                    if " BendThick" in item:
                        row_split = item.split()
                        prop_BendThick = float(row_split[1])
                        parsed[prop_id].append(prop_BendThick)
                    if item[0] == '/':
                        break
        if len(parsed) == 0:
            return None
        return parsed

    @staticmethod
    def __parse_model_static_nl_combo(path: Path) -> None | List[str]:
        parsed = []
        read_block = False
        with open(path) as file:
            for item in file:
                if not read_block:
                    if item[0] == '/':
                        if "/ NON-LINEAR STATIC SOLVER DATA" in item:
                            read_block = True
                else:
                    if "NonLinearIncrement" in item:
                        row_splitted = item.split()
                        combo_name = str(row_splitted[4].strip('"'))
                        parsed.append(combo_name)
                    if item[0] == '/':
                        break
        return parsed

    @staticmethod
    def __make_file_assignments_for_rebar(
            path: Path,
            groups: np.ndarray,
            ll: Literal[0, 1, 2, 3]
    ) -> bool:
        df_rows_nb = len(np.unique(groups))
        rebar_colum = np.full(df_rows_nb, dtype=object, fill_value="[[nb=, diam=, dist=, from=]]")
        shear_colum = np.full(df_rows_nb, dtype=object, fill_value="[[nb=, diam=, dist=, step=]]")
        rebar_mat_code = np.full(df_rows_nb, dtype=object, fill_value="NTC2008")
        rebar_mat_key = np.full(df_rows_nb, dtype=object, fill_value="B450C")
        rebar_mat_sensitivity = np.full(df_rows_nb, dtype=object, fill_value="not sensitive")
        concrete_mat_code = np.full(df_rows_nb, dtype=object, fill_value="NTC2008")
        concrete_mat_key = np.full(df_rows_nb, dtype=object, fill_value="NTC2008")
        concrete_mat_environment = np.full(df_rows_nb, dtype=object, fill_value="not aggressive")
        thick_reduced = np.full(df_rows_nb, dtype=float, fill_value=0.0)

        combined = np.vstack((
            groups,
            rebar_colum,
            rebar_colum,
            shear_colum,
            rebar_mat_code,
            rebar_mat_key,
            rebar_mat_sensitivity,
            concrete_mat_code,
            concrete_mat_key,
            concrete_mat_environment,
            thick_reduced
        )).transpose()
        rebars_assignments = pd.DataFrame(
            combined,
            columns=[
                "group",
                "lay13_rebar",
                "lay24_rebar",
                "shear_rebar",
                "rebar_code",
                "rebar_key",
                "rebar_sensitivity",
                "concrete_code",
                "concrete_key",
                "concrete_environment",
                "thick_reduced"
            ]
        )
        rebars_assignments.to_excel(path)
        msg = f"File {path.name} ... created !!!"
        log("INF", msg, ll, True)
        return True

    @staticmethod
    def __make_file_assignments_for_check(
            path: Path,
            combo: List[str],
            ll: Literal[0, 1, 2, 3]
    ) -> bool:
        df_rows_nb = len(combo)
        check_colum = np.full(df_rows_nb, dtype=object, fill_value="X")
        no_check_colum = np.full(df_rows_nb, dtype=object, fill_value="")
        combined = np.vstack((
            np.array(combo),
            check_colum,
            check_colum,
            check_colum,
            check_colum,
            no_check_colum,
            no_check_colum,
            no_check_colum,
            no_check_colum,
            no_check_colum,
            no_check_colum
        )).transpose()
        check_assignments = pd.DataFrame(
            combined,
            columns=["combo_name",
                     "CHECK_SLU_NM",
                     "CHECK_SLU_T",
                     "CHECK_SLE_NM",
                     "CHECK_SLE_F",
                     "SLU",
                     "SLV",
                     "SLE_R",
                     "SLE_F",
                     "SLE_QP",
                     "ACC"
                     ]
        )
        check_assignments.to_excel(path)
        msg = f"File {path.name} ... created !!!"
        log("INF", msg, ll, True)
        return True

    @staticmethod
    def __parse_assignment_for_rebar(path: Path, ll: Literal[0, 1, 2, 3]) -> None | dict:
        msg = f"File {path.name} already exists. I'm reading ..."
        log("INF", msg, ll, True)
        columns = (
            "group",
            "lay13_rebar",
            "lay24_rebar",
            "shear_rebar",
            "rebar_code",
            "rebar_key",
            "rebar_sensitivity",
            "concrete_code",
            "concrete_key",
            "concrete_environment",
            "thick_reduced"
        )
        assignments_df = pd.read_excel(path, usecols=columns)
        assignments = dict()
        for r in assignments_df.to_numpy():
            lay13_rebar_assigned = None
            lay24_rebar_assigned = None
            shear_rebar_assigned = None
            try:
                lay13_rebar_assigned = eval(r[1])
            except SyntaxError:
                msg = f"SyntaxError in lay13 rebar assign at group *{r[0]}* !!!"
                log("ERR", msg, ll, True)
            except NameError:
                msg = f"NameError in lay13 rebar assign at group *{r[0]}* !!!"
                log("ERR", msg, ll, True)
            try:
                lay24_rebar_assigned = eval(r[2])
            except SyntaxError:
                msg = f"SyntaxError in lay24 rebar assign at group *{r[0]}* !!!"
                log("ERR", msg, ll, True)
            except NameError:
                msg = f"NameError in lay24 rebar assign at group *{r[0]}* !!!"
                log("ERR", msg, ll, True)
            try:
                shear_rebar_assigned = eval(r[3])
            except SyntaxError:
                msg = f"SyntaxError in shear rebar assign at group *{r[0]}* !!!"
                log("ERR", msg, ll, True)
            except NameError:
                msg = f"NameError in shear rebar assign at group *{r[0]}* !!!"
                log("ERR", msg, ll, True)
            c1 = lay13_rebar_assigned is not None
            c2 = lay24_rebar_assigned is not None
            c3 = shear_rebar_assigned is not None
            if all([c1, c2, c3]):
                assignments[r[0]] = [
                    lay13_rebar_assigned,
                    lay24_rebar_assigned,
                    shear_rebar_assigned,
                    r[4], r[5], r[6], r[7], r[8], r[9], r[10]
                ]
            else:
                return None

        return assignments

    @staticmethod
    def __parse_assignment_for_check(path: Path, ll: Literal[0, 1, 2, 3]) -> None | dict:
        msg = f"File {path.name} already exists. I'm reading ..."
        log("INF", msg, ll, True)
        columns = (
            "combo_name",
            "CHECK_SLU_NM",
            "CHECK_SLU_T",
            "CHECK_SLE_NM",
            "CHECK_SLE_F",
            "SLU",
            "SLV",
            "SLE_R",
            "SLE_F",
            "SLE_QP",
            "ACC"
        )
        assignments_df = pd.read_excel(path, usecols=columns)
        assignments = {}
        for r in assignments_df.to_numpy():
            SLU_NM = True if r[1] == 'X' else False
            SLU_T = True if r[2] == 'X' else False
            SLE_NM = True if r[3] == 'X' else False
            SLE_F = True if r[4] == 'X' else False
            combo_type = ""
            if r[5] == 'X':
                combo_type = "SLU"
            if r[6] == 'X':
                combo_type = "SLV"
            if r[7] == 'X':
                combo_type = "SLE_R"
            if r[8] == 'X':
                combo_type = "SLE_F"
            if r[9] == 'X':
                combo_type = "SLE_QP"
            if r[10] == 'X':
                combo_type = "ACC"
            assignments[r[0]] = {
                "combo_name": r[0],
                "SLU_NM": SLU_NM,
                "SLU_T": SLU_T,
                "SLE_NM": SLE_NM,
                "SLE_F": SLE_F,
                "combo_type": combo_type
            }
        if len(assignments) == 0:
            assignments = None
        return assignments

    @staticmethod
    def __extract_plate_layer_combo(res: ArrayLike):
        plate_id = int(res[0].split(':')[0].split(' ')[1])
        layer_id = 0
        if len(res[0].split(':')) == 2:
            layer_id = int(res[0].split(':')[1].split(' ')[2])
        combo_id = int(res[1].split(':')[0])
        combo_name = res[1].split(':')[1].split(' ')[1]
        return plate_id, layer_id, combo_id, combo_name

    def __checker_for_plate(
            self,
            plate_id: int,
            layer_id: int,
            combo_id: int,
            combo_name: str,
            plate_thick: float,
            F1: float,
            F2: float,
            rebar_assignment: list,
            check_assignment: dict,
            group_name: str,
            print_nook: bool = True,
            save_fig_nook: bool = False,
            save_fig_ok: bool = False,
            check_required: str = 'SLU_NM',
            ll: Literal[0, 1, 2, 3] = 3
    ) -> dict:
        check_result = dict()
        check_result['global_check'] = True
        code_steel = Code(rebar_assignment[3])
        steel = ConcreteSteel()
        steel.setByCode(code_steel, rebar_assignment[4])
        steel.setSensitivity(rebar_assignment[5])

        code_concrete = Code(rebar_assignment[6])
        concrete = Concrete()
        concrete.setByCode(code_concrete, rebar_assignment[7])
        concrete.setEnvironment(rebar_assignment[8])

        KN = 1000
        KNm = 1000 * 1000
        m = 1000

        section = RCTemplRectEC2(1, f"Template RC Section for plate {plate_id} with thick = {plate_thick*1000:.0f}")
        section.setConcreteMaterial(concrete)
        section.setSteelMaterial(steel)
        section.setDimH(plate_thick * m - rebar_assignment[9])
        section.setDimW(1000)

        err = False
        force = Load()
        Med = 0.0
        Ned = 0.0
        Ted = 0.0
        if check_assignment['combo_type'] == 'SLU' and check_assignment['SLU_NM'] and  check_required == 'SLU_NM':
            Med = F1
            Ned = F2
            if layer_id == 3 or layer_id == 4:
                Med = -Med
            # Normal efforts must be multiplied for 2 cause Wood Armer Moment
            force.id = combo_id
            force.Fx = 2*Ned * KN
            force.My = Med * KNm
            force.descr=f"Layer: {layer_id} Combo: {combo_name}"
            force.limitState = LimiteState_Enum.ULTIMATE
            check_result['SLU_NM'] = dict()
        elif check_assignment['combo_type'] == 'SLU' and check_assignment['SLU_T']  and  check_required == 'SLU_T':
            Ted = float(np.sqrt(F1**2 + F2**2))
            force.id = combo_id
            force.Fz = Ted * KN
            force.descr=f"Combo: {combo_name}"
            force.limitState = LimiteState_Enum.ULTIMATE
            check_result['SLU_T'] = dict()
        elif check_assignment['combo_type'] == 'SLE_R' and check_assignment['SLE_NM']  and  check_required == 'SLE_NM':
            Med = F1
            Ned = F2
            if layer_id == 3 or layer_id == 4:
                Med = -Med
            # Normal efforts must be multiplied for 2 cause Wood Armer Moment
            force.id=combo_id
            force.Fx = 2*Ned * KN
            force.My = Med * KNm
            force.descr=f"Layer: {layer_id} Combo: {combo_name}"
            force.limitState = LimiteState_Enum.SERVICEABILITY
            force.frequency = Frequency_Enum.CHARACTERISTIC
            check_result['SLE_NM'] = dict()
        elif check_assignment['combo_type'] == 'SLE_QP' and check_assignment['SLE_NM']  and  check_required == 'SLE_NM':
            Med = F1
            Ned = F2
            if layer_id == 3 or layer_id == 4:
                Med = -Med
            # Normal efforts must be multiplied for 2 cause Wood Armer Moment
            force.id=combo_id
            force.Fx = 2*Ned * KN
            force.My = Med * KNm
            force.descr=f"Layer: {layer_id} Combo: {combo_name}"
            force.limitState = LimiteState_Enum.SERVICEABILITY
            force.frequency = Frequency_Enum.QUASI_PERMANENT
            check_result['SLE_NM'] = dict()
        elif check_assignment['combo_type'] == 'SLE_F' and check_assignment['SLE_F']  and  check_required == 'SLE_F':
            Med = F1
            Ned = F2
            if layer_id == 3 or layer_id == 4:
                Med = -Med
            # Normal efforts must be multiplied for 2 cause Wood Armer Moment
            force.id=combo_id
            force.Fx = 2*Ned * KN
            force.My = Med * KNm
            force.descr=f"Layer: {layer_id} Combo: {combo_name}"
            force.limitState = LimiteState_Enum.SERVICEABILITY
            force.frequency = Frequency_Enum.FREQUENT
            check_result['SLE_F'] = dict()
        elif check_assignment['combo_type'] == 'SLE_QP' and check_assignment['SLE_F']  and  check_required == 'SLE_F':
            Med = F1
            Ned = F2
            if layer_id == 3 or layer_id == 4:
                Med = -Med
            # Normal efforts must be multiplied for 2 cause Wood Armer Moment
            force.id=combo_id
            force.Fx = 2*Ned * KN
            force.My = Med * KNm
            force.descr=f"Layer: {layer_id} Combo: {combo_name}"
            force.limitState = LimiteState_Enum.SERVICEABILITY
            force.frequency = Frequency_Enum.QUASI_PERMANENT
            check_result['SLE_F'] = dict()
        else:
            err = True
            msg = f"None check for SLU and SLU_NM selection !!!"
            log("WRN", msg, ll, True)

        if not err:

            if layer_id == 1 or layer_id == 3:
                rebars_aligned_arr = rebar_assignment[0]
            elif layer_id == 2 or layer_id == 4:
                rebars_aligned_arr = rebar_assignment[1]
            else:
                rebars_aligned_arr = []

            # We need this because height section util for shear for example
            if layer_id == 0:
                rebars_aligned_arr = rebar_assignment[0] + rebar_assignment[1]

            for r in rebars_aligned_arr:
                if r[3] < 0:
                    section.addSteelArea("LINE-MT", dist=-r[3], d=r[1], nb=r[0], sd=r[2])
                else:
                    section.addSteelArea("LINE-MB", dist=r[3], d=r[1], nb=r[0], sd=r[2])

            checkable = RcsRectangular(section)
            perform_save_fig = False
            if check_required == 'SLU_NM':
                checkable.setMaxPointsForSLUDomain(20)
                res = checkable.check_SLU_NM_NTC2018(force)
                check_result['SLU_NM'] = res
                check_result['safetyFactor'] = res['safetyFactor']['interactionDomain']
                if res['safetyFactor']['interactionDomain'] < 1:
                    check_result['global_check'] = False
                    if print_nook:
                        msg = (f"Safety factor is {res['safetyFactor']['interactionDomain']} for plate "
                               f"{plate_id} layer {layer_id} combo_name {combo_name}")
                        log("ERR", msg, ll, True)
                        msg = f"       : Group Name: {group_name} Thickness is {section.getDimH():.0f}mm"
                        log("WRN", msg, ll, True)
                        msg = f"       : Med = {Med} Ned = {2*Ned} Rebar: {rebars_aligned_arr}"
                        log("WRN", msg, ll, True)
                    if save_fig_nook:
                        perform_save_fig = True
                else:
                    if save_fig_ok:
                        perform_save_fig = True

                if perform_save_fig:
                    gn = group_name.replace('\\', '-')
                    fn = (self.__working_path /
                          self.__out_rel_path /
                          Path(f"group-{gn}-plate-{plate_id}-layer-{layer_id}-combo-{combo_name}.png"))
                    section.interactionDomainPlot2d(
                        xLabel="N [KN]",
                        yLabel="M [KN*m]",
                        export=str(fn)
                    )

            elif check_required == 'SLU_T':
                stirrup_arr = rebar_assignment[2]
                sum_Awi = 0
                sum_si = 0
                for r in stirrup_arr:
                    nb_leg = r[0]
                    diam_leg = r[1]
                    area_leg = diam_leg**2 * np.pi / 4
                    sum_Awi += area_leg * nb_leg
                    sum_si += r[3]
                average_si = sum_si / len(stirrup_arr)
                shear_area = sum_Awi
                section.setStirrupt(area = shear_area, dist = average_si, angle = 90)
                res = checkable.check_SLU_T_NTC2018(force)
                check_result['SLU_T'] = res
                check_result['safetyFactor'] = None
                if res["safetyFactor"] is not None:
                    check_result['safetyFactor'] = res['safetyFactor']['globalCheck']
                    if check_result['safetyFactor'] < 1:
                        check_result['global_check'] = False
                        if print_nook:
                            msg = (f"WARNING: Safety factor is {check_result['safetyFactor']} for plate {plate_id}"
                                   f"layer {layer_id} combo_name {combo_name} ")
                            log("WRN", msg, ll, True)
                            msg = f"       : Group Name: {group_name} Thickness is {section.getDimH():.0f}"
                            log("WRN", msg, ll, True)
                            msg = f"       : Ted = {Ted} Shear Area: {shear_area} Average Step: {average_si}"
                            log("WRN", msg, ll, True)

            elif check_required == 'SLE_NM':
                res = checkable.check_SLE_NM_NTC2018(force)
                check_result['SLE_NM'] = res
                check_result['global_check'] = res["check"]["globalCheck"]
                check_result['safetyFactor'] = None
                # Is None Type when section is totally stretched
                if res["safetyFactor"]["globalCheck"] is not None:
                    check_result['safetyFactor'] = res["safetyFactor"]["globalCheck"]
                    if check_result['safetyFactor'] < 1:
                        if print_nook:
                            msg = (f"WARNING: Safety factor is {check_result['safetyFactor']} for plate {plate_id} "
                                   f"layer {layer_id} combo_name {combo_name} ")
                            log("WRN", msg, ll, True)
                            msg = f"       : Group Name: {group_name} Thickness is {section.getDimH():.0f}"
                            log("WRN", msg, ll, True)
                            msg = f"       : Med = {Med} Ned = {2*Ned} Rebar: {rebars_aligned_arr}"
                            log("WRN", msg, ll, True)

            elif check_required == 'SLE_F':
                res = checkable.check_SLE_F_NTC2018(force)
                check_result['SLE_F'] = res
                check_result['global_check'] = res["check"]["crack"]
                check_result['safetyFactor'] = None
                # Is None Type when section is totally stretched
                if res["safetyFactor"]["crack"] is not None:
                    check_result['safetyFactor'] = res["safetyFactor"]["crack"]
                    if check_result['safetyFactor'] < 1:
                        if print_nook:
                            msg = (f"WARNING: Safety factor is {check_result['safetyFactor']} for plate {plate_id} "
                                   f"layer {layer_id} combo_name {combo_name} ")
                            log("WRN", msg, ll, True)
                            msg = f"       : Group Name: {group_name} Thickness is {section.getDimH():.0f}"
                            log("WRN", msg, ll, True)
                            msg = f"       : Med = {Med} Ned = {2*Ned} Rebar: {rebars_aligned_arr}"
                            log("WRN", msg, ll, True)
        return check_result

    @staticmethod
    def __write_to_excel_with_row_limited(
            file_path: Path,
            rows_per_file: int,
            data: Union[None, NDArray[Any]],
            ll: Literal[0, 1, 2, 3]
    ):
        number_of_files = math.floor((len(data) / rows_per_file)) + 1
        start_index = 0
        end_index = rows_per_file
        columns = ["plate_nb", "layer", "combo_name", "combo_id", "plate_thick", "F1", "F2", "rebar_properties",
                   "check_properties", "group", "check_performed", "safety_factor", "plate_X", "plate_Y", "plate_Z"]
        df = pd.DataFrame(data, columns=columns)
        log("INF", f" ... data shape originale {df.shape}", ll, True)

        for i in range(number_of_files):
            filepart = Path(str(file_path) + str(Path('_file_' + str(i) + '.xlsx')))
            log("INF", f" ... writing file {filepart} {df.shape}", ll, True)

            with pd.ExcelWriter(filepart, engine='openpyxl') as writer:
                df_mod = df.iloc[start_index:end_index]
                log("INF", f" ... data shape chunked {df_mod.shape}", ll, True)
                df_mod.to_excel(writer, index=False, sheet_name='sheet')
                start_index = end_index
                end_index = end_index + rows_per_file
                df_mod.to_excel(writer)

    def export_checks_to_files(self, file_path: Path, write_json=False, export_nook=True, export_xlsx=True):
        path = self.__working_path / self.__out_rel_path / file_path
        stream_in = self.__plate_results_stream
        stream_out = self.__plate_results_log_stream
        nook_indices = self.__idx_check_nook

        # Write json verbose and slow process
        if write_json:
            log("INF", f"I'm writing json ... ", self.__ll, True)
            with open(str(path) + ".json", 'w') as f:
                json.dump(stream_out, f, indent=4)
            log("INF", f"... done", self.__ll, True)



        # Export NOOK results
        #
        if len(nook_indices) > 0:
            log("INF", f"I risultati con check negativo sono {len(nook_indices)}", self.__ll, True)

            plate_results_stream_nook = np.zeros((len(nook_indices), len(stream_in[0])), dtype=object)

            for idx, val in enumerate(nook_indices):
                plate_results_stream_nook[idx] = np.array(stream_in[val], dtype=object)

            if export_nook:
                log("INF", f"I'm writing nook results on excel file ... ", self.__ll, True)
                self.__write_to_excel_with_row_limited(
                    Path(str(path) + "_nook"),
                    50000,
                    plate_results_stream_nook,
                    self.__ll
                )
                log("INF", f"... done", self.__ll, True)

        else:
            log("INF", f"Nook results are zero lenght. Can't export it.", self.__ll, True)

        if export_xlsx:
            log("INF", f"I'm writing results on excel file ... ", self.__ll, True)
            self.__write_to_excel_with_row_limited(
                path,
                50000,
                stream_in,
                self.__ll
            )
            log("INF", f"... done", self.__ll, True)
        else:
            log("INF", f"Required don't export massive excel results files.", self.__ll, True)

    def results_checker(
            self,
            max_row: int | None = None,
            check_required: str = 'SLU_NM',
            save_fig_nook: bool = False,
            save_fig_ok: bool = False
        ) -> bool:
        results = self.__results_parsed
        row_to_check = len(results)
        elements_2D_properties = self.__elements_2D_properties
        check_assignments = self.__assignments_for_check
        rebar_assignments = self.__assignments_for_rebar
        if max_row is None:
            plate_results_stream = [dict()] * row_to_check
            plate_results_log_stream = [dict()] * row_to_check
        else:
            plate_results_stream = [dict()] * min(row_to_check, max_row)
            plate_results_log_stream = [dict()] * min(row_to_check, max_row)

        idx = 0
        msg = f"Number of results for check {row_to_check}"
        log("WRN", msg, self.__ll, True)
        idx_check_nook = []
        start_single = time.time()
        min_safety_factor = 1000000000000000
        max_safety_factor = -1000000000000000

        for res in results:
            plate_id, layer_id, combo_id, combo_name = self.__extract_plate_layer_combo(res)

            # Filter on Elements dataframe
            #
            if self.__filter_on_group is not None:
               if elements_2D_properties.get(plate_id) is None:
                    idx += 1
                    continue

            property_id = elements_2D_properties[plate_id][2]
            group_name = elements_2D_properties[plate_id][1]

            elements_XX = elements_2D_properties[plate_id][5]
            elements_YY = elements_2D_properties[plate_id][6]
            elements_ZZ = elements_2D_properties[plate_id][7]

            try:
                rebar_for_check = rebar_assignments[group_name]
            except KeyError:
                log("ERR", f"Error in {group_name} fetching rebar assignment", self.__ll, True)
                log("ERR", f"Probably you didn't check Show Group Path in Strand7 while exporting.", self.__ll, True)
                return False

            plate_thick = self.__model_plate_properties[property_id][1]
            plate_input_row = (plate_id, layer_id, combo_id, combo_name, plate_thick, res[2], res[3], rebar_for_check,
                               check_assignments[combo_name], group_name)
            output_stream = self.__checker_for_plate(
                *plate_input_row,
                save_fig_nook=save_fig_nook,
                check_required=check_required,
                save_fig_ok=save_fig_ok
            )
            plate_results_log_stream[idx] = output_stream

            plate_results_stream[idx] = plate_input_row + (check_required, output_stream['safetyFactor'], elements_XX,
                                                           elements_YY, elements_ZZ)

            if plate_results_log_stream[idx]['safetyFactor'] is not None:
                if plate_results_log_stream[idx]['safetyFactor'] < min_safety_factor:
                    min_safety_factor = plate_results_log_stream[idx]['safetyFactor']

                if plate_results_log_stream[idx]['safetyFactor'] > max_safety_factor:
                    max_safety_factor = plate_results_log_stream[idx]['safetyFactor']

            # Build index for nook check result
            if not plate_results_log_stream[idx]['global_check']:
                idx_check_nook.append(idx)

            nb_elements_trace = 1000
            if idx / nb_elements_trace == np.floor(idx / nb_elements_trace):
                end_single = time.time()
                elapsed = end_single - start_single
                remaining = (len(results) - idx) * (elapsed / nb_elements_trace)
                msg = (f"Elaborati {idx} di {row_to_check} - {idx / row_to_check * 100:.4f} % in {elapsed:.2f} sec. "
                       f"Restano {remaining:.0f} sec. (min sf: {min_safety_factor:.2f} - max sf: {max_safety_factor:.2f})")
                log("INF", msg, self.__ll, True)
                start_single = time.time()

            idx += 1
            if max_row is not None:
                if idx == max_row:
                    break
        log("INF", f"Minimo fattore di sicurezza è {min_safety_factor}", self.__ll, True)
        log("INF", f"Massimo fattore di sicurezza è {max_safety_factor}", self.__ll, True)

        self.__plate_results_stream = plate_results_stream
        self.__plate_results_log_stream = plate_results_log_stream
        self.__idx_check_nook = idx_check_nook

        if len(plate_results_stream) > 0:
            # TODO: change "_" with "-" and remove <.replace('_','-')> from line below
            self.__curr_check_required = Crit(check_required.replace('_','-'))
            self.__results_checked[self.__curr_input_results_name] = {
                self.__curr_input_results_tp: {
                    self.__curr_check_required: plate_results_stream
                }
            }
        return True

    def getModelPlateProperties(self):
        return self.__model_plate_properties

    def getModelStaticNLCombo(self):
        return self.__model_static_nl_combo

    def getAssignmentsForRebars(self):
        return self.__assignments_for_rebar

    def getAssignmentsForCheck(self):
        return self.__assignments_for_check

    def getPlateResultsStream(self):
        return self.__plate_results_stream

    def getPlateResultsLogStream(self):
        return self.__plate_results_log_stream

    def getCheckNookIndices(self):
        return self.__idx_check_nook

    def getElement2DProperties(self):
        return self.__elements_2D_properties

    def parse(self) -> bool:

        if self.__working_path is None:
            msg = f"Working path is none. Assign it with setWorkingPath() !!!"
            log("ERR", msg, self.__ll, True)
            return False

        if not self.__working_path.exists():
            msg = (f"Working path {self.__working_path} doesn't exists. "
                   f"Assign it with setWorkingPath() !!!")
            log("ERR", msg, self.__ll, True)
            return False

        in_file_path = self.__working_path / self.__in_rel_path
        if not in_file_path.exists():
            msg = (f"Input path {self.__in_rel_path} in working path "
                   f"{self.__working_path} doesn't exists. You have to create it !!!")
            log("ERR", msg, self.__ll, True)
            return False

        out_file_path = self.__working_path / self.__out_rel_path
        if not out_file_path.exists():
            msg = (f"Output path {self.__out_rel_path} in working path "
                   f"{self.__working_path} doesn't exists. You have to create it !!!")
            log("ERR", msg, self.__ll, True)
            return False

        # Parsing of NODES
        #
        nodes_file_path = self.__working_path / self.__in_rel_path / Path(self.__fn_nodes)
        if not nodes_file_path.exists():
            msg = (f"File of nodes {nodes_file_path} doesn't exists. "
                   f"Export it with Strand7 or change path !!!")
            log("ERR", msg, self.__ll, True)
            return False
        self.__nodes_df = pd.read_csv(nodes_file_path, header=[0, 1])
        self.__nodes_id = self.__nodes_df.iloc[:, [0]].to_numpy()
        self.__nodes_coords = self.__nodes_df.iloc[:, [2, 3, 4]].to_numpy()
        msg = f"File of nodes {nodes_file_path} done."
        log("INF", msg, self.__ll, True)

        # Parsing of ELEMENTS
        #
        element_file_path = self.__working_path / self.__in_rel_path / Path(self.__fn_elements)
        if not element_file_path.exists():
            msg = (f"File of elements {element_file_path} doesn't exists. "
                   f"Export it with Strand7 or change path !!!")
            log("ERR", msg, self.__ll, True)
            return False

        self.__elements_df = pd.read_csv(element_file_path, header=[0, 1])
        log("INF", f"Elements read from file {len(self.__elements_df)}", self.__ll, True)
        if self.__filter_on_group is not None:
            mask_on_group = self.__elements_df["Group"].isin(self.__filter_on_group)
            self.__elements_df = self.__elements_df[mask_on_group.values]
            log("INF", f"Elements filtered {len(self.__elements_df)}", self.__ll, True)

        self.__elements_2D_id = self.__elements_df.loc[:, "Plate"].to_numpy()
        self.__elements_2D_groups = self.__elements_df.loc[:, "Group"].to_numpy()

        # Forming arrays for PYVISTA
        #
        elements_2D_connectivity = \
            self.__elements_df.iloc[:, [14, 15, 16, 17, 18, 19, 20, 21, 22]].fillna(0).to_numpy()
        self.__elements_2D_connectivity = []
        self.__elements_2D_type = []
        nb_lines = 0
        nb_triangles = 0
        nb_quad = 0

        for e in elements_2D_connectivity:
            element = [len(np.trim_zeros(e)), *list(np.trim_zeros(e))]
            indexes = [len(np.trim_zeros(e))]

            for i in element[1:len(element)]:
                indexes.append(np.where(self.__nodes_id == i)[0][0])

            self.__elements_2D_connectivity += indexes
            if element[0] == 2:
                self.__elements_2D_type.append(pv.CellType.LINE)
                nb_lines += 1
            if element[0] == 3:
                self.__elements_2D_type.append(pv.CellType.TRIANGLE)
                nb_triangles += 1
            if element[0] == 4:
                self.__elements_2D_type.append(pv.CellType.QUAD)
                nb_quad += 1

        log("INF", f"Parsed {nb_lines} lines | {nb_triangles} triangles | {nb_quad} quads", self.__ll, True)

        # Forming __elements_2D_properties
        #
        elements_property = self.__elements_df.loc[:, "Property"].to_numpy()
        elements_angle = self.__elements_df.loc[:, "Angle"].to_numpy()
        elements_X = self.__elements_df.loc[:, "X"].to_numpy()
        elements_Y = self.__elements_df.loc[:, "Y"].to_numpy()
        elements_Z = self.__elements_df.loc[:, "Z"].to_numpy()
        self.__elements_2D_properties = dict()
        for i, v in np.ndenumerate(self.__elements_2D_id):
            prop_id = int(elements_property[i[0]][0].split(':')[0])
            prop_str = elements_property[i[0]][0].split(':')[1].strip()
            self.__elements_2D_properties[v] = [
                i[0],
                self.__elements_2D_groups[i[0]][0],
                prop_id,
                prop_str,
                elements_angle[i[0]][0],
                elements_X[i[0]][0],
                elements_Y[i[0]][0],
                elements_Z[i[0]][0]
            ]
        # Parsing or creating REBAR ASSIGNMENTS file
        #
        # Build a dataframe for external rebar data formed by:
        #  -------------------------------------------------------------------------------
        # |  group  |  lay1_rebar  |  lay3_rebar | lay2_rebar | lay4_rebar | shear_legs |
        #  -------------------------------------------------------------------------------
        # Example for lay1_rebar:
        # -----------------------
        #              [[4, 20, 100, -50], [8, 26, 80, +50]]
        # means:
        # two rebars lines with 4 rebar d=20mm with distance 100mm between rebars and 50mm
        # from top (- sign) + 8 rebar d=26mm with distance 80mm  between rebars and 50mm
        # from bottom (+ sign)
        #
        # Example for shear_legs
        # -----------------------
        #             [[4, 12, 100, 250], [4, 10, 150, 200]]
        # means:
        # 4 legs diam=12mm with transverse distance 100mm and longitudinal step 250mm
        #
        rebars_file_path = self.__working_path / self.__in_rel_path / Path(self.__fn_assignments_for_rebar)
        if not rebars_file_path.exists():
            msg = (f"File of rebar assignments {rebars_file_path} doesn't exists. "
                   f"Will be created. Check it out and compile it !!!")
            log("INF", msg, self.__ll, True)
            self.__make_file_assignments_for_rebar(
                rebars_file_path,
                np.unique(self.__elements_2D_groups),
                self.__ll
            )
            return False
        else:
            self.__assignments_for_rebar = self.__parse_assignment_for_rebar(
                rebars_file_path,
                self.__ll
            )
            if self.__assignments_for_rebar is None:
                msg = f"... error reading !!! Quit."
                log("ERR", msg, self.__ll, True)
                return False
            msg = f"... read successfully !!!"
            log("INF", msg, self.__ll, True)

        # Parsing MODEL for plate properties
        model_file_path = self.__working_path / self.__in_rel_path / Path(self.__fn_model)
        if not model_file_path.exists():
            msg = (f"Model file {model_file_path} doesn't exists. "
                   f"Export it with Strand7 or change path !!!")
            log("ERR", msg, self.__ll, True)
            self.__model_plate_properties = None
            self.__model_static_nl_combo = None
            return False

        self.__model_plate_properties = self.__parse_model_plate_properties(model_file_path)
        if self.__model_plate_properties is None:
            msg = f"Model file {model_file_path} parsed plate properties with error."
            log("ERR", msg, self.__ll, True)
            return False
        msg = f"Model file {model_file_path} parsed plate properties with success."
        log("INF", msg, self.__ll, True)

        self.__model_static_nl_combo = self.__parse_model_static_nl_combo(model_file_path)
        if len(self.__model_static_nl_combo) == 0:
            msg = f"Model file {model_file_path} parsed plate properties with error."
            log("ERR", msg, self.__ll, True)
            return False
        msg = f"Model file {model_file_path} parsed not linear combinations with success."
        log("INF", msg, self.__ll, True)

        # Parsing or creating CHECK ASSIGNMENTS file
        #
        rebars_check_file_path = self.__working_path / self.__in_rel_path / Path(self.__fn_assignments_for_check)
        if not rebars_check_file_path.exists():
            msg = (f"File of check assignments {rebars_check_file_path} doesn't exists. "
                   f"Will be created. Check it out and compile it !!!")
            log("INF", msg, self.__ll, True)
            self.__make_file_assignments_for_check(
                rebars_check_file_path,
                self.__model_static_nl_combo,
                self.__ll
            )
            return False
        else:
            self.__assignments_for_check = self.__parse_assignment_for_check(
                rebars_check_file_path,
                self.__ll
            )
            if self.__assignments_for_rebar is None:
                msg = f"... error reading !!! Quit."
                log("ERR", msg, self.__ll, True)
                return False
            msg = f"... read successfully !!!"
            log("INF", msg, self.__ll, True)
        return True

    def setWorkingPath(self, working_path: str) -> bool:
        path = Path(working_path)

        if not path.exists():
            msg = (f"Working path {path} doesn't exists. "
                   f"Assign it with setWorkingPath() !!!")
            log("ERR", msg, self.__ll, True)
            return False

        self.__working_path = path
        log("INF", f"Working path <{self.__working_path}> done.", self.__ll, True)
        return True

    def readResults(self, file_name: Path, file_type: Literal["WA","FORCE"], skip_nb_rows: int = 0) -> bool:
        path = self.__working_path / self.__in_rel_path / file_name
        self.__curr_input_results_name = str(file_name)
        if file_type == "WA":
            self.__curr_input_results_tp = ResultsType.WA
            results_parsed = pd.read_csv(path, header=[0], skiprows=[1, 2] + [i for i in range(3, 3 + skip_nb_rows)])
            self.__results_parsed = results_parsed.to_numpy()
        elif file_type == "FORCE":
            self.__curr_input_results_tp = ResultsType.FORCE
            results_parsed = pd.read_csv(path, header=[0], skiprows=[1] + [i for i in range(2, 2 + skip_nb_rows)])
            self.__results_parsed = results_parsed.to_numpy()[:, [0, 1, 5, 6]]
        else:
            log("ERR", f"You have to choose WA or FORCE type", self.__ll, True)
            return False

        log("INF", f"You can check max {len(self.__results_parsed)} results", self.__ll, True)
        return True

    def plot_model(self, off_screen=False) -> Plotter:
        grid = pv.UnstructuredGrid(
            self.__elements_2D_connectivity,
            self.__elements_2D_type,
            self.__nodes_coords
        )
        pl = pv.Plotter(off_screen=off_screen)
        pl.show_axes()
        pl.add_mesh(grid, show_edges=True, line_width=1)
        return pl

    def plot_figure(self, results_fn: str, results_tp: str, check_required: str, skip_less_than: float | None = None) -> plt:
        results_list = self.__results_checked[results_fn][ResultsType(results_tp)][Crit(check_required.replace('_','-'))]
        results_arr = np.zeros((len(results_list),2))
        results_arr_skipped = np.zeros((len(results_list), 2))
        skipped = False
        nb_skipped = 0
        for i, r in enumerate(results_list):
            if type(r[12]) is np.float64:
                if skip_less_than is not None:
                    if r[11] <= skip_less_than:
                        skipped = True
                        nb_skipped += 1
                        results_arr_skipped[i] = [r[11], r[12]]
                    else:
                        results_arr[i] = [r[11],r[12]]
                else:
                    results_arr[i] = [r[11], r[12]]
            else:
                skipped = True
                nb_skipped += 1
                results_arr_skipped[i] = [r[11],r[12]]

        if skipped:
            log("WRN", f"Nb. {nb_skipped} results was skipped from {len(results_arr)}!!!", self.__ll, True)

        # results_arr = np.array(results_list)
        idx_col_sf = 0
        idx_col_xx = 1

        results_arr_no_zero = results_arr[results_arr[:,idx_col_sf] != 0]

        plt.figure()
        # plt.subplot(211)
        plt.xlabel('Progressive lenght [m]')
        plt.ylabel('Safety Factor (< 1 means check ok)')
        plt.title(f"File name: '{results_fn}' \n File type: '{results_tp}' - Check Required: '{check_required}'")
        plt.plot(results_arr_no_zero[:,idx_col_xx], results_arr_no_zero[:,idx_col_sf]**(-1), 'bo')
        path_where_save = self.__working_path / self.__out_rel_path
        fn_without_suffix = str(path_where_save / Path(results_fn).stem) + f"-{results_tp}" + f"-{check_required}"
        self.__path_last_figure_plotted = Path(fn_without_suffix).with_suffix(".png")
        plt.savefig(self.__path_last_figure_plotted)
        log("INF", f"Plotting figure {fn_without_suffix} ... ", self.__ll, True)
        plt.savefig(fn_without_suffix, dpi = 300)
        log("INF", f"... plotted.", self.__ll, True)
        return plt



